# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import unicode_literals

import codecs
import json
import logging
import os
from os import path
import textwrap

from jinja2 import Environment

log = logging.getLogger(__name__)

TMPL_TAG = """
{%- for tag in swagger.tags -%}

.. swagger:tag:: {{tag.name}}
   :synopsis: {{tag.description}}
{% for line in tag['x-summary'].split('\n') %}
   {{line}}
{%- endfor %}

{% endfor %}
"""

TMPL_API = """
{%- for path, methods in swagger['paths'].items() -%}
{%- for method_name, request in methods.items() -%}

{%- if method_name.startswith('x-') -%}
.. http:{{ method_name|remove_prefix(request['operationId']) }}:: {{path}}
{%- else %}
.. http:{{method_name}}:: {{path}}
{%- endif %}
   :title: {{request['x-title']}}
   :synopsis: {{request['summary']}}
{%- if request.description != '' %}
{% for line in request.description.split('\n') %}
   {{line}}
{%- endfor %}
{%- endif %}
{% for status_code, response in request.responses.items() %}
{%- if response['examples']['application/json'] %}
   :responseexample {{status_code}}: {{version}}/examples/{{request['operationId']}}_resp_{{status_code}}.json
{%- endif -%}
{%- if response['examples']['text/plain'] %}
   :responseexample {{status_code}}: {{version}}/examples/{{request['operationId']}}_resp_{{status_code}}.txt
{%- endif -%}
{%- if response['schema']['$ref'] %}
   :responseschema {{status_code}}: {{version}}/{{response['schema']['$ref'].rsplit('/', 1)[1]}}.json
{%- endif -%}
{% endfor -%}
{% for mime in request.consumes %}
   :accepts: {{mime}}
{%- endfor -%}
{% for mime in request.produces %}
   :produces: {{mime}}
{%- endfor -%}
{% for tag in request.tags %}
   :tag: {{tag}}
{%- endfor -%}
{% for parameter in request.parameters -%}
{% if parameter.in == 'body' -%}
{% if parameter.schema %}
   :requestschema: {{version}}/{{request['operationId']}}.json
{%- for id, schema in swagger['definitions'].items() -%}
{%- if id == request['operationId'] -%}
{%- if 'example' in schema -%}
{%- if schema['example']['application/json'] %}
   :requestexample: {{version}}/examples/{{request['operationId']}}_req.json
{%- endif -%}
{%- if schema['example']['text/plain'] %}
   :requestexample: {{version}}/examples/{{request['operationId']}}_req.txt
{%- endif -%}
{%- endif -%}
{%- endif -%}
{%- endfor -%}
{%- endif -%}
{%- elif parameter.in == 'path' %}
{{ parameter|format_param('parameter') }}
{%- elif parameter.in == 'query' %}
{{ parameter|format_param('query') }}
{%- elif parameter.in == 'header' %}
{{ parameter|format_param('reqheader') }}
{%- endif %}
{%- endfor -%}
{% for status_code, response in request['responses'].items() %}
   :statuscode {{status_code}}: {{response.description}}
{%- endfor %}


{% endfor %}
{%- endfor %}
"""  # noqa

environment = Environment()


def format_param(obj, type='query'):
    param = '   :%s %s: ' % (type, obj['name'])
    param_wrap = textwrap.TextWrapper(
        initial_indent=param,
        subsequent_indent=' ' * len(param))
    if ('::\n\n' in obj['description']) or \
       ('.. code-block::' in obj['description']):
        param_len = len(param)
        for i, line in enumerate(obj['description'].split('\n')):
            if i is 0:
                param += line + '\n'
            else:
                param += ' ' * param_len + line + '\n'
        return param
    else:
        new_text = param_wrap.wrap(obj['description'])
        return '\n'.join(new_text)


def remove_prefix(obj, op_name):
    if obj[2:].endswith(op_name):
        return obj[2:(len(obj) - len(op_name) - 1)]
    else:
        return obj[2:]

environment.filters['format_param'] = format_param
environment.filters['remove_prefix'] = remove_prefix


def main1(filename, output_dir):
    log.info('Parsing %s' % filename)
    swagger = json.load(open(filename))
    write_rst(swagger, output_dir)
    write_jsonschema(swagger, output_dir)
    write_examples(swagger, output_dir)
    write_index(swagger, output_dir)


def write_index(swagger, output_dir):
    info = swagger['info']
    service = info['x-service']

    # Web UI uses 'service' field for indexing
    del info['x-service']
    info['service'] = service
    version = info['version']
    output_file = 'index.json'
    filepath = path.join(output_dir, output_file)
    log.info("Writing APIs %s", filepath)
    if path.exists(filepath):
        index = json.load(open(filepath))
    else:
        index = {}
    index['/'.join([service, version, ''])] = info
    with codecs.open(filepath,
                     'w', "utf-8") as out_file:
        json.dump(index, out_file, indent=2)


def write_rst(swagger, output_dir):
    environment.extend(swagger_info=swagger['info'])
    write_apis(swagger, output_dir)
    write_tags(swagger, output_dir)


def write_apis(swagger, output_dir):
    info = swagger['info']
    version = info['version']
    service = info['x-service']
    service_path = path.join(output_dir, service)
    output_file = '%s.rst' % version
    if not path.exists(service_path):
        os.makedirs(service_path)
    TMPL = environment.from_string(TMPL_API)
    result = TMPL.render(swagger=swagger,
                         version=swagger['info']['version'])
    filepath = path.join(service_path, output_file)
    log.info("Writing APIs %s", filepath)
    with codecs.open(filepath,
                     'w', "utf-8") as out_file:
        out_file.write(result)


def write_tags(swagger, output_dir):
    info = swagger['info']
    version = info['version']
    service = info['x-service']
    service_path = path.join(output_dir, service)
    if not path.exists(service_path):
        os.makedirs(service_path)
    output_file = '%s-tags.rst' % version
    TMPL = environment.from_string(TMPL_TAG)
    result = TMPL.render(swagger=swagger,
                         version=swagger['info']['version'])
    filepath = path.join(service_path, output_file)
    log.info("Writing Tags %s", filepath)
    with codecs.open(filepath,
                     'w', "utf-8") as out_file:
        out_file.write(result)


def write_jsonschema(swagger, output_dir):
    info = swagger['info']
    version = info['version']
    service = info['x-service']
    service_path = path.join(output_dir, service)
    full_path = path.join(service_path, version)
    if not path.exists(service_path):
        os.makedirs(service_path)
    if not path.exists(full_path):
        os.makedirs(full_path)

    for schema_name, schema in swagger['definitions'].items():
        filename = '%s.json' % schema_name
        filepath = path.join(full_path, filename)
        log.info("Writing %s", filepath)
        file = open(filepath, 'w')
        json.dump(schema, file, indent=2)


def write_examples(swagger, output_dir):
    info = swagger['info']
    version = info['version']
    service = info['x-service']
    service_path = path.join(output_dir, service)
    versioned_path = path.join(service_path, version)
    full_path = path.join(versioned_path, 'examples')
    if not path.exists(service_path):
        os.makedirs(service_path)
    if not path.exists(versioned_path):
        os.makedirs(versioned_path)
    if not path.exists(full_path):
        os.makedirs(full_path)

    for paths in swagger['paths'].values():
        for operation in paths.values():
            for status_code, response in operation['responses'].items():
                for mime, example in response['examples'].items():
                    filename = '%s' % '_'.join([operation['operationId'],
                                                'resp',
                                                status_code])
                    if mime == 'application/json':
                        filepath = path.join(full_path, filename + '.json')
                        log.info("Writing %s", filepath)
                        file = open(filepath, 'w')
                        json.dump(example, file, indent=2)
                    if mime == 'text/plain':
                        filepath = path.join(full_path, filename + '.txt')
                        log.info("Writing %s", filepath)
                        example = example.strip()
                        example = example + '\n'
                        file = open(filepath, 'w')
                        file.write(example)

    for ids, schemas in swagger['definitions'].items():
        if 'example' in schemas:
            for mime, example in schemas['example'].items():
                filename = '%s' % '_'.join(
                    [ids, 'req'])
                if mime == 'application/json':
                    filepath = path.join(full_path, filename + '.json')
                    log.info("Writing %s", filepath)
                    file = open(filepath, 'w')
                    json.dump(example, file, indent=2)
                if mime == 'text/plain':
                    filepath = path.join(full_path, filename + '.txt')
                    log.info("Writing %s", filepath)
                    example = example.strip()
                    example = example + '\n'
                    file = open(filepath, 'w')
                    file.write(example)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-v', '--verbose', action='count', default=0,
        help="Increase verbosity (specify multiple times for more)")
    parser.add_argument(
        '-o', '--output-dir', action='store',
        help="The directory to output the JSON files too.")
    parser.add_argument(
        'filename',
        help="File to convert")

    args = parser.parse_args()

    log_level = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG

    logging.basicConfig(
        level=log_level,
        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    filename = path.abspath(args.filename)

    main1(filename, output_dir=args.output_dir)



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import logging
from os import path

import docutils.core
from pecan import conf
from pecan import expose
from pecan import response
from webob.exc import status_map
from webob.static import FileIter

from fairy_slipper import hooks
from fairy_slipper.rest import JSONWriter

logger = logging.getLogger(__name__)


class JSONFileController(object):

    __hooks__ = [hooks.CORSHook()]

    def __init__(self, filepath):
        self.filepath = filepath

    @expose('json')
    @expose(content_type='text/plain')
    def _default(self):
        if path.exists(self.filepath + '.json'):
            self.filepath = self.filepath + '.json'
        if path.exists(self.filepath + '.txt'):
            self.filepath = self.filepath + '.txt'
        if not path.exists(self.filepath):
            response.status = 404
            return response
        response.app_iter = FileIter(open(self.filepath, 'rb'))
        return response


class DocController(object):

    __hooks__ = [hooks.CORSHook()]

    def __init__(self, service_path, service_info):
        self.service_info = service_info
        base_filepath = path.join(conf.app.api_doc, service_path.rstrip('/'))
        self.api_rst = base_filepath + '.rst'
        self.tags_rst = base_filepath + '-tags.rst'
        self.examples_dir = path.join(base_filepath, 'examples') + path.sep
        self.schema_dir = base_filepath + path.sep
        if not path.exists(self.api_rst):
            logger.warning("Can't find ReST API doc at %s", self.api_rst)
        if not path.exists(self.tags_rst):
            logger.warning("Can't find ReST TAG doc at %s", self.tags_rst)

    @expose('json')
    def index(self):
        if path.exists(self.tags_rst) and path.exists(self.api_rst):
            rst = open(self.api_rst).read() + \
                "\n\n" + open(self.tags_rst).read()
        elif path.exists(self.api_rst):
            rst = open(self.api_rst).read()
        else:
            logger.warning("Can't find ReST documents to render.")
            return {}

        json = docutils.core.publish_string(rst, writer=JSONWriter())

        return {'info': self.service_info,
                'paths': json['paths'],
                'tags': json['tags']}

    @expose()
    def _lookup(self, *components):
        if len(components) != 2 and len(components) != 3:
            return

        if components[0] == 'examples':
            example = components[1]
            filepath = path.join(self.examples_dir, example)
            return JSONFileController(filepath), []
        else:
            filename = components[0]
            print(filename)
            filepath = path.join(self.schema_dir, filename)
            return JSONFileController(filepath), []


class ServicesController(object):

    def __init__(self):
        filepath = path.join(conf.app.api_doc, 'index.json')
        self.url_map = {}
        if not path.exists(filepath):
            logger.error("Can't find documentation at %s", filepath)
            self.services_info = {}
            return
        try:
            self.services_info = json.load(open(filepath))
        except ValueError:
            logger.error("Failed to load %s", filepath)
            raise
        for key, info in self.services_info.items():
            # Add the path into each element, this is to make
            # consumption by the JS client easier.
            info['url'] = key

            current_map = self.url_map
            previous_map = None
            for part in [k for k in key.split('/') if k]:
                if part not in current_map:
                    current_map[part] = {}
                previous_map = current_map
                current_map = current_map[part]
            else:
                previous_map[part] = DocController(key, info)

    @expose('json')
    def index(self):
        return list(self.services_info.values())

    @expose('json')
    def _lookup(self, *components):
        url_map = self.url_map
        url_walk = itertools.chain(components)
        for component in url_walk:
            if component in url_map:
                url_map = url_map[component]
            else:
                break
            if isinstance(url_map, DocController):
                return url_map, [u for u in url_walk]


class RootController(object):

    def __init__(self):
        self.doc = ServicesController()

    @expose(content_type='text/html')
    def index(self):
        filepath = path.join(conf.app.static_root, 'index.html')
        f = open(filepath, 'rb')
        response.app_iter = FileIter(f)

    @expose('error.html')
    def error(self, status):
        try:
            status = int(status)
        except ValueError:
            status = 0
        message = getattr(status_map.get(status), 'explanation', '')
        return dict(status=status, message=message)



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import unittest
import xml.sax

from fairy_slipper.cmd import wadl_to_swagger


class MockParent(object):
    result = None
    rest = None

    def detach_subparser(self, result, **kwargs):
        self.result = result
        self.kwargs = kwargs


class TestParaParser(unittest.TestCase):

    def test_code_block(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>This is an example request:</para>
  <programlisting>GET /v2.0/routers/{router_id}
Accept: application/json</programlisting>
  <para>para2</para>
</wadl:doc>
"""

        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """This is an example request:

::

   GET /v2.0/routers/{router_id}
   Accept: application/json

para2

""")

    def test_code_block_language(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>This is an example request:</para>
<programlisting language="json">"OS-OAUTH1": {
    "access_token_id": "cce0b8be7"
}</programlisting>
  <para>para2</para>
</wadl:doc>
"""

        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """This is an example request:

.. code-block:: json

   "OS-OAUTH1": {
       "access_token_id": "cce0b8be7"
   }

para2

""")

    def test_link(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>
    To create a keypair, make a <link
    xlink:href="http://developer.openstack.org/#createKeypair">
    create keypair</link> request.
  </para>
</wadl:doc>
"""

        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            "To create a keypair, make a "
            "`create keypair\n"
            "<http://developer.openstack.org/#createKeypair>`_"
            " request.\n\n")

    def test_anonymous_link(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>
    To create a keypair, see <link
    xlink:href="http://developer.openstack.org/#createKeypair"></link> this link.
  </para>
</wadl:doc>
"""  # noqa

        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            "To create a keypair, see\n"
            "`<http://developer.openstack.org/#createKeypair>`__"
            " this link.\n\n")

    def test_para_inline_code(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>Code sample <code>admin</code></para>
  <para><code>get task</code>, program!</para>
  <para><code>get task</code> program!</para>
  <para>Code sample <code>admin</code> and more code.</para>
  <para>Code <code>admin</code>.</para>
  <para>Code <code>admin</code>. Another sentence started.</para>
  <para>para5</para>
  <para>Code <code>test</code>: with colon</para>
  <para>Code <code>test</code>; with semi-colon</para>
  <para>Code <code>test</code>; with semi-colon and <code>
  end test</code>.</para>
  <para>Code (<code>test</code>) with parens.</para>
</wadl:doc>
"""

        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """Code sample ``admin``

``get task``, program!

``get task`` program!

Code sample ``admin`` and more code.

Code ``admin``.

Code ``admin``. Another sentence started.

para5

Code ``test``: with colon

Code ``test``; with semi-colon

Code ``test``; with semi-colon and ``end test``.

Code (``test``) with parens.

""")

    def test_para_emphasis(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>Some text with <emphasis>bold words at the end of the sentence</emphasis>.</para>
  <para><emphasis>Bold text is cool</emphasis></para>
  <para>Multi word text is <emphasis>brilliant</emphasis>and then some.</para>
  <para>Really <emphasis>bold</emphasis></para>
  <para>para5</para>
  <para>This is a <emphasis>sentence</emphasis>; another phrase follows.</para>
  <para>Text with (<emphasis>parenthesis</emphasis>) and then **text**: with colon.</para>
  <para>A <emphasis>comma</emphasis>, and more <code>text</code>.</para>
  <para>Some text with <emphasis>back</emphasis> <emphasis>to back</emphasis>
        emphasized text.</para>
</wadl:doc>
"""  # noqa

        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """Some text with **bold words at the end of the sentence**.

**Bold text is cool**

Multi word text is **brilliant** and then some.

Really **bold**

para5

This is a **sentence**; another phrase follows.

Text with (**parenthesis**) and then **text**: with colon.

A **comma**, and more ``text``.

Some text with **back** **to back** emphasized text.

""")

    def test_listitem_para(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>This operation does not accept a request body.</para>
  <para>Example requests and responses:</para>
  <itemizedlist>
     <listitem><para>Show account details and list containers:</para>
     <para>Some more details.</para>
     </listitem>
     <listitem><para>See the example response below.</para></listitem>
  </itemizedlist>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """This operation does not accept a request body.
\nExample requests and responses:
\n- Show account details and list containers:
\n  Some more details.
\n- See the example response below.\n\n"""
        )

    def test_nested_listitem(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
<itemizedlist>
<listitem>
<para>Para 1, listitem1</para>
<para>Para 2, listitem1</para>
<itemizedlist>
<listitem>
<para>Embedded item1</para>
</listitem>
<listitem>
<para>Embedded item2</para>
</listitem>
<listitem>
<para>Embedded item3</para>
</listitem>
</itemizedlist>
<para>Para 3, listitem1</para>
</listitem>
<listitem><para>Para1, listitem2</para></listitem>
</itemizedlist>
<para>Another para of text</para>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """- Para 1, listitem1\n\n  Para 2, listitem1\n\n  - Embedded item1
\n  - Embedded item2\n\n  - Embedded item3\n\n  Para 3, listitem1
\n- Para1, listitem2\n\nAnother para of text\n\n"""
        )

    def test_listitem_para_code(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
   <para>This operation does not accept a request body.</para>
   <para>Example requests and responses:</para>
   <itemizedlist>
      <listitem><para>Show account details and list containers:</para>
                <para><code>curl -i
                            $publicURL?format=json -X GET -H
                            "X-Auth-Token: $token"</code></para>
                <para>See the example response below.</para>
      </listitem>
      </itemizedlist>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """This operation does not accept a request body.
\nExample requests and responses:
\n- Show account details and list containers:
\n  ``curl -i $publicURL?format=json -X GET -H \"X-Auth-Token:\n  $token\"``
\n  See the example response below.\n\n"""
        )

    def test_listitem_para_programlisting(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
<para>Delete the <code>steven</code>container:</para>
<itemizedlist>
<listitem>
<para>Container command:</para>
<para>
<code>curl -i $publicURL/steven
 -X DELETE -H "X-Auth-Token: $token"</code>
</para>
<para>If the container does not exist, the response is:</para>
<para><programlisting>HTTP/1.1 404 Not Found
Content-Length: 70
Content-Type: text/html; charset=UTF-8
Date: Thu, 16 Jan 2014 18:00:20 GMT
&lt;html>
&lt;h1>Conflict&lt;/h1>
&lt;p>Trying to complete your request.&lt;/p>&lt;/html>
</programlisting></para>
</listitem>
<listitem><para>Second container command:</para><para>Write to disk.</para>
</listitem>
</itemizedlist>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """Delete the ``steven`` container:
\n- Container command:
\n  ``curl -i $publicURL/steven -X DELETE -H \"X-Auth-Token: $token\"``
\n  If the container does not exist, the response is:
\n  ::\n\n     HTTP/1.1 404 Not Found\n     Content-Length: 70
     Content-Type: text/html; charset=UTF-8
     Date: Thu, 16 Jan 2014 18:00:20 GMT
     <html>\n     <h1>Conflict\n     </h1>
     <p>Trying to complete your request.\n     </p>
     </html>\n\n\n- Second container command:
\n  Write to disk.\n\n"""
        )

    def test_para_code_block(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>This is an example request:</para>
  <para><programlisting>GET /v2.0/routers/{router_id}
Accept: application/json</programlisting></para>
  <para>para2</para>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """This is an example request:

::

   GET /v2.0/routers/{router_id}
   Accept: application/json


para2

""")

    def test_para_code_block_language(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <para>This is an example request:</para>
<para><programlisting language="json">"OS-OAUTH1": {
    "access_token_id": "cce0b8be7"
}</programlisting></para>
  <para>para2</para>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """This is an example request:

.. code-block:: json

   "OS-OAUTH1": {
       "access_token_id": "cce0b8be7"
   }


para2

""")

    def test_listitem_all_in_one_para(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
        <para>Example requests and responses:</para>
            <itemizedlist>
                <listitem><para>Copy the <code>goodbye</code> object
                        from the <code>marktwain</code> container to
                        the <code>janeausten</code> container:
                            <code>curl -i $publicURL/marktwain/goodbye
                            -X COPY -H "X-Auth-Token: $token" -H
                            "Destination: janeausten/goodbye"</code>
                        <programlisting>HTTP/1.1 201 Created
Content-Length: 0
X-Copied-From: marktwain/goodbye
</programlisting></para></listitem>
</itemizedlist>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """Example requests and responses:
\n- Copy the ``goodbye`` object from the ``marktwain`` container to
  the ``janeausten`` container: ``curl -i
  $publicURL/marktwain/goodbye -X COPY -H "X-Auth-Token: $token" -H
  "Destination: janeausten/goodbye"``  ::\n\n     HTTP/1.1 201 Created
     Content-Length: 0\n     X-Copied-From: marktwain/goodbye\n\n\n"""
        )

    def test_table_caption(self):
        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<wadl:doc>
  <table><caption>Image status</caption></table>
  <table><caption>Image <code>with a code literal</code>
  <code>inside</code></caption></table>
  <table>
  <caption>A <emphasis>bold</emphasis> caption <emphasis>again</emphasis>
  </caption></table>
  <table>
  <caption>
  <emphasis role="italic">An italicized</emphasis> caption</caption>
  </table>
  <table>
  <caption>A caption with <emphasis>bold</emphasis> text embedded</caption>
  </table>
</wadl:doc>
"""
        parent = MockParent()
        ch = wadl_to_swagger.ParaParser(parent)
        xml.sax.parse(StringIO(file_content), ch)
        self.assertEqual(
            parent.result,
            """**Image status**

++

**Image with a code literal inside**

++

**A bold caption again**

++

**An italicized caption**

++

**A caption with bold text embedded**

++

""")


class TestWADLHandler(unittest.TestCase):

    def test_simple_wadl(self):
        filename = "api-v1.wadl"
        api_ref = {
            "file_tags": {filename: 'things'},
            "method_tags": {
            },
            "resource_tags": {
            },
            "service": "lorem",
            "tags": [],
            "title": "Lorem Ipsum",
            "version": "v1"
        }

        file_content = """<?xml version="1.0" encoding="UTF-8"?>
<application>
  <resources xml:id="os-attach-v2">
    <resource id="version" type="#VersionDetails" path="v2/">
      <method href="#createThing" />
    </resource>
  </resources>
  <method name="POST" id="createThing">
    <wadl:doc title="Create interface">
      <para role="shortdesc">Creates and uses a port interface
        to attach the port to a server instance.</para>
    </wadl:doc>
    <request>
      <representation mediaType="application/json">
        <param name="thing-a-imagig" style="plain"
               type="xsd:string" required="true">
          <wadl:doc>
            <para>
               Specify the <code>interfaceAttachment</code>
               action in the request body.
            </para>
          </wadl:doc>
        </param>
      </representation>
    </request>
    <response status="202">
      <representation mediaType="application/json">
        <param name="thing-a-imagig-response" style="plain"
               type="xsd:string" required="true">
          <wadl:doc>
            <para>
               Specify the <code>interfaceAttachment</code>
               action in the request body.
            </para>
          </wadl:doc>
        </param>
      </representation>
    </response>
  </method>
</application>
        """

        ch = wadl_to_swagger.WADLHandler(filename, api_ref)
        xml.sax.parse(StringIO(file_content), ch)

        self.assertEqual(
            ch.apis,
            {'v2/':
             [{'consumes': [],
               'description': '',
               'examples': {},
               'operationId': 'createThing',
               'method': 'post',
               'parameters': [
                   {'description': '',
                    'in': 'body',
                    'name': 'body',
                    'required': False,
                    'schema': {'$ref':
                               '#/definitions/createThing'}}],
               'produces': [],
               'responses': {'202': {'examples': {},
                                     'headers': {},
                                     'schema': {
                                         '$ref':
                                         '#/definitions/createThing_202'},
                                     'description': ''}},
               'summary': 'Creates and uses a port interface '
               'to attach the port to a server instance.',
               'tags': ['things'],
               'title': 'Create interface'}]})

        self.assertEqual(
            ch.schemas,
            {'createThing':
             {'properties':
              {'thing-a-imagig':
               {'description':
                'Specify the ``interfaceAttachment``'
                ' action in the request body.',
                'format': '',
                'required': True,
                'type': 'string'}},
              'type': 'object'},
             'createThing_202':
             {'properties':
              {'thing-a-imagig-response':
               {'description':
                'Specify the ``interfaceAttachment``'
                ' action in the request body.',
                'format': '',
                'required': True,
                'type': 'string'}},
              'type': 'object'}}
        )



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pecan import make_app


def setup_app(config):

    app_conf = dict(config.app)

    return make_app(
        app_conf.pop('root'),
        logging=getattr(config, 'logging', {}),
        **app_conf
    )



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import unittest

from fairy_slipper.cmd import tempest_log


SIMPLE_LOG = """2015-09-04 15:51:29.023 18793 DEBUG tempest_lib.common.rest_client [req-30784c0a-e9a1-4411-a7c1-20715b26598f ] Request (FlavorsV2TestJSON:setUpClass): 200 POST http://192.168.122.201:5000/v2.0/tokens
2015-09-04 15:51:29.023 18793 DEBUG tempest_lib.common.rest_client [req-30784c0a-e9a1-4411-a7c1-20715b26598f ] Request - Headers: {}
        Body: None
    Response - Headers: {'status': '200', 'content-length': '2987', 'vary': 'X-Auth-Token', 'server': 'Apache/2.4.7 (Ubuntu)', 'connection': 'close', 'date': 'Sun, 13 Sep 2015 07:43:01 GMT', 'content-type': 'application/json', 'x-openstack-request-id': 'req-1'}
        Body: None
2015-09-04 15:51:45.472 18793 INFO tempest_lib.common.rest_client [req-b710aeba-6263-4a49-bf50-2da42227c870 ] Request (FlavorsV2TestJSON:test_get_flavor): 200 POST http://192.168.122.201:5000/v2.0/tokens
2015-09-04 15:51:45.472 18793 DEBUG tempest_lib.common.rest_client [req-b710aeba-6263-4a49-bf50-2da42227c870 ] Request - Headers: {}
        Body: None
    Response - Headers: {'status': '200', 'content-length': '2987', 'vary': 'X-Auth-Token', 'server': 'Apache/2.4.7 (Ubuntu)', 'connection': 'close', 'date': 'Sun, 13 Sep 2015 07:43:01 GMT', 'content-type': 'application/json', 'x-openstack-request-id': 'req-2'}
        Body: None
"""  # noqa

SIMPLE_LOG_BODY = """2015-09-04 15:51:29.007 18793 INFO tempest_lib.common.rest_client [req-9e329507-e0ce-448c-a363-f49e39dd96b0 ] Request (FlavorsV2TestJSON:test_get_flavor): 200 GET http://192.168.122.201:8774/v2.1/6b45254f6f7c44a1b65ddb8218932226/flavors/1 0.117s
2015-09-04 15:51:29.007 18793 DEBUG tempest_lib.common.rest_client [req-9e329507-e0ce-448c-a363-f49e39dd96b0 ] Request - Headers: {'Content-Type': 'application/json', 'Accept': 'application/json', 'X-Auth-Token': '<omitted>'}
        Body: None
    Response - Headers: {'status': '200', 'content-length': '430', 'content-location': 'http://192.168.122.201:8774/v2.1/6b45254f6f7c44a1b65ddb8218932226/flavors/1', 'x-compute-request-id': 'req-959a09e8-3628-419d-964a-1be4ca604232', 'vary': 'X-OpenStack-Nova-API-Version', 'connection': 'close', 'x-openstack-nova-api-version': '2.1', 'date': 'Sun, 13 Sep 2015 07:43:01 GMT', 'content-type': 'application/json'}
        Body: {"flavor": {"name": "m1.tiny", "links": [{"href": "http://192.168.122.201:8774/v2.1/6b45254f6f7c44a1b65ddb8218932226/flavors/1", "rel": "self"}, {"href": "http://192.168.122.201:8774/6b45254f6f7c44a1b65ddb8218932226/flavors/1", "rel": "bookmark"}], "ram": 512, "OS-FLV-DISABLED:disabled": false, "vcpus": 1, "swap": "", "os-flavor-access:is_public": true, "rxtx_factor": 1.0, "OS-FLV-EXT-DATA:ephemeral": 0, "disk": 1, "id": "1"}}
"""  # noqa

DEBUG_LOG = """2015-09-04 15:54:42.296 18793 INFO tempest_lib.common.rest_client [req-39c6042e-5a4a-4517-9fe9-32b34cfaa5a8 ] Request (TestSessionsTenantIsolation:test_delete_session_in_env_from_another_tenant): 403 DELETE http://127.0.0.1:8082/v1/environments/7501923609b145ec88eeb4a5c93e371c/sessions/db214e36e0494c4e9dc67fb0df8548f7 0.010s
2015-09-04 15:54:42.296 18793 DEBUG tempest_lib.common.rest_client [req-39c6042e-5a4a-4517-9fe9-32b34cfaa5a8 ] Request - Headers: {'Content-Type': 'application/json', 'Accept': 'application/json', 'X-Auth-Token': '<omitted>'}
        Body: None
    Response - Headers: {'status': '403', 'content-length': '75', 'connection': 'close', 'date': 'Fri, 04 Sep 2015 15:54:42 GMT', 'content-type': 'text/plain; charset=UTF-8', 'x-openstack-request-id': 'req-39c6042e-5a4a-4517-9fe9-32b34cfaa5a8'}
        Body: 403 Forbidden

User is not authorized to access these tenant resources

    _log_request_full /opt/stack/new/tempest/.venv/local/lib/python2.7/site-packages/tempest_lib/common/rest_client.py:411
2015-09-04 15:52:13.727 18793 INFO tempest_lib.common.rest_client [req-0ff36a16-dacd-49c8-9835-7ce92d50f5a7 ] Request (TestEnvironmentsTenantIsolation:tearDown): 200 DELETE http://127.0.0.1:8082/v1/environments/c32c6d5095c4476da549ed065e9b5196 0.054s
2015-09-04 15:52:13.727 18793 DEBUG tempest_lib.common.rest_client [req-0ff36a16-dacd-49c8-9835-7ce92d50f5a7 ] Request - Headers: {'Content-Type': 'application/json', 'Accept': 'application/json', 'X-Auth-Token': '<omitted>'}
        Body: None
    Response - Headers: {'status': '200', 'content-length': '0', 'connection': 'close', 'date': 'Fri, 04 Sep 2015 15:52:13 GMT', 'content-type': 'application/json', 'x-openstack-request-id': 'req-0ff36a16-dacd-49c8-9835-7ce92d50f5a7'}
        Body:  _log_request_full /opt/stack/new/tempest/.venv/local/lib/python2.7/site-packages/tempest_lib/common/rest_client.py:411
"""  # noqa


DEBUG_LOG_AUTH = """2015-09-04 15:49:46.056 14923 INFO tempest_lib.common.rest_client [req-280bc347-e650-473e-92bb-bcc59103e12c ] Request (main): 200 POST http://127.0.0.1:5000/v2.0/tokens
2015-09-04 15:49:46.056 14923 DEBUG tempest_lib.common.rest_client [req-280bc347-e650-473e-92bb-bcc59103e12c ] Request - Headers: {}
        Body: None
    Response - Headers: {'server': 'Apache/2.4.7 (Ubuntu)', 'vary': 'X-Auth-Token', 'x-openstack-request-id': 'req-280bc347-e650-473e-92bb-bcc59103e12c', 'content-length': '4846', 'connection': 'close', 'status': '200', 'content-type': 'application/json', 'date': 'Fri, 04 Sep 2015 15:49:42 GMT'}
        Body: None _log_request_full /opt/stack/new/tempest/.tox/venv/local/lib/python2.7/site-packages/tempest_lib/common/rest_client.py:411
"""  # noqa


def db_to_call_list(db):
    calls = []
    for req in sorted(db.requests):
        calls.append((db.requests[req], db.responses[req]))
    return calls


class TestLogParser(unittest.TestCase):
    maxDiff = 10000

    def test_simple_parse(self):
        result = db_to_call_list(
            tempest_log.parse_logfile(StringIO(SIMPLE_LOG)))
        self.assertEqual(result, [
            ({'url': '/v2.0/tokens',
              'service': 'identity',
              'headers': {},
              'body': None,
              'method': 'POST'},
             {'status_code': '200',
              'body': None,
              'headers': {'status': '200',
                          'content-length': '0',
                          'date': 'Sun, 13 Sep 2015 07:43:01 GMT',
                          'content-type': 'application/json',
                          'x-openstack-request-id': 'req-1',
                          'vary': 'X-Auth-Token',
                          'connection': 'close',
                          'server': 'Apache/2.4.7 (Ubuntu)'}}),
            ({'url': '/v2.0/tokens',
              'service': 'identity',
              'headers': {},
              'body': None,
              'method': 'POST'},
             {'status_code': '200',
              'body': None,
              'headers': {'status': '200',
                          'content-length': '0',
                          'date': 'Sun, 13 Sep 2015 07:43:01 GMT',
                          'content-type': 'application/json',
                          'x-openstack-request-id': 'req-2',
                          'vary': 'X-Auth-Token',
                          'connection': 'close',
                          'server': 'Apache/2.4.7 (Ubuntu)'}})])

    def test_body_parse(self):
        result = db_to_call_list(
            tempest_log.parse_logfile(StringIO(SIMPLE_LOG_BODY)))

        self.assertEqual(result, [
            ({'url': '/v2.1/6b45254f6f7c44a1b65ddb8218932226/flavors/1',
              'headers': {'content-type': 'application/json',
                          'content-length': '0',
                          'accept': 'application/json',
                          'x-auth-token': '<omitted>'},
              'body': None,
              'method': 'GET',
              'service': 'compute'},
             {'body': '{\n  "flavor": {\n    "OS-FLV-DISABLED:disabled": false,\n    "OS-FLV-EXT-DATA:ephemeral": 0,\n    "disk": 1,\n    "id": "1",\n    "links": [\n      {\n        "href": "http://192.168.122.201:8774/v2.1/6b45254f6f7c44a1b65ddb8218932226/flavors/1",\n        "rel": "self"\n      },\n      {\n        "href": "http://192.168.122.201:8774/6b45254f6f7c44a1b65ddb8218932226/flavors/1",\n        "rel": "bookmark"\n      }\n    ],\n    "name": "m1.tiny",\n    "os-flavor-access:is_public": true,\n    "ram": 512,\n    "rxtx_factor": 1.0,\n    "swap": "",\n    "vcpus": 1\n  }\n}',  # noqa
              'status_code': '200',
              'headers': {'status': '200', 'content-length': '548',
                          'content-location': 'http://192.168.122.201:8774/v2.1/6b45254f6f7c44a1b65ddb8218932226/flavors/1',  # noqa
                          'x-openstack-nova-api-version': '2.1',
                          'date': 'Sun, 13 Sep 2015 07:43:01 GMT',
                          'vary': 'X-OpenStack-Nova-API-Version',
                          'x-compute-request-id': 'req-959a09e8-3628-419d-964a-1be4ca604232',  # noqa
                          'content-type': 'application/json',
                          'connection': 'close'}})])

    def test_debug_log(self):
        result = db_to_call_list(
            tempest_log.parse_logfile(StringIO(DEBUG_LOG)))

        self.assertEqual(result, [
            ({'body': None,
              'headers': {'accept': 'application/json',
                          'content-type': 'application/json',
                          'content-length': '0',
                          'x-auth-token': '<omitted>'},
              'method': 'DELETE',
              'service': 'application-catalog',
              'url': '/v1/environments/c32c6d5095c4476da549ed065e9b5196'},
             {'body': None,
              'headers': {'connection': 'close',
                          'content-length': '0',
                          'content-type': 'application/json',
                          'date': 'Fri, 04 Sep 2015 15:52:13 GMT',
                          'status': '200',
                          'x-openstack-request-id':
                          'req-0ff36a16-dacd-49c8-9835-7ce92d50f5a7'},
              'status_code': '200'}),
            ({'body': None,
              'headers': {'accept': 'application/json',
                          'content-type': 'application/json',
                          'content-length': '0',
                          'x-auth-token': '<omitted>'},
              'method': 'DELETE',
              'service': 'application-catalog',
              'url': '/v1/environments/7501923609b145ec88eeb4a5c93e371c'
              '/sessions/db214e36e0494c4e9dc67fb0df8548f7'},
             {'body': '403 Forbidden\n'
              'User is not authorized to access these tenant resources\n\n',
              'headers': {'connection': 'close',
                          'content-length': '13',
                          'content-type': 'text/plain; charset=UTF-8',
                          'date': 'Fri, 04 Sep 2015 15:54:42 GMT',
                          'status': '403',
                          'x-openstack-request-id':
                          'req-39c6042e-5a4a-4517-9fe9-32b34cfaa5a8'},
              'status_code': '403'})])

    def test_debug_admin_log(self):
        result = db_to_call_list(
            tempest_log.parse_logfile(StringIO(DEBUG_LOG_AUTH)))

        self.assertEqual(result, [
            ({'body': None,
              'headers': {},
              'method': 'POST',
              'service': 'identity',
              'url': '/v2.0/tokens'},
             {'body': None,
              'headers': {'connection': 'close',
                          'content-length': '0',
                          'content-type': 'application/json',
                          'date': 'Fri, 04 Sep 2015 15:49:42 GMT',
                          'server': 'Apache/2.4.7 (Ubuntu)',
                          'status': '200',
                          'vary': 'X-Auth-Token',
                          'x-openstack-request-id':
                          'req-280bc347-e650-473e-92bb-bcc59103e12c'},
              'status_code': '200'})])



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pecan.hooks import PecanHook


class CORSHook(PecanHook):
    def after(self, state):
        state.response.headers['Access-Control-Allow-Origin'] = '*'
        state.response.headers['Access-Control-Allow-Methods'] = 'GET'
        state.response.headers['Access-Control-Allow-Headers'] = \
            'origin, authorization, accept'



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from fairy_slipper.tests import FunctionalTest


class TestRootControllerNegative(FunctionalTest):

    def setUp(self):
        self.CONFIG['app']['api_doc'] = \
            '%(confdir)s/fairy_slipper/tests/api_doc_missing'
        super(TestRootControllerNegative, self).setUp()

    def test_get_doc_index(self):
        response = self.app.get('/doc/')
        assert response.json == []
        assert response.status_int == 200


class TestRootController(FunctionalTest):

    def test_get(self):
        response = self.app.get('/')
        assert response.status_int == 200

    def test_search(self):
        response = self.app.post('/', params={'q': 'RestController'})
        assert response.status_int == 200

    def test_get_not_found(self):
        response = self.app.get('/a/bogus/url', expect_errors=True)
        assert response.status_int == 404

    def test_get_doc_index(self):
        response = self.app.get('/doc/')
        assert response.json == [
            {'url': 'identity/v2/',
             'version': 'v2',
             'license': {
                 'url': 'http://www.apache.org/licenses/LICENSE-2.0.html',
                 'name': 'Apache 2.0'},
             'service': 'identity',
             'title': 'Identity'}]
        assert response.status_int == 200

    def test_get_doc_identity(self):
        response = self.app.get('/doc/identity/', expect_errors=True)
        assert response.status_int == 404

    def test_get_doc_identity_v2(self):
        response = self.app.get('/doc/identity/v2/')
        assert response.json == \
            {'info':
             {'url': 'identity/v2/',
              'version': 'v2',
              'license': {
                  'url': 'http://www.apache.org/licenses/LICENSE-2.0.html',
                  'name': 'Apache 2.0'},
              'service': 'identity',
              'title': 'Identity'},
             'paths': {'/': [{'responses': {},
                              'parameters': [],
                              'produces': [],
                              'consumes': [],
                              'tags': ['simple'],
                              'summary': '',
                              'title': 'Simple route',
                              'method': 'get',
                              'description': ''}]},
             'tags': [{'description': '',
                       'name': 'simple',
                       'summary': 'Simple Tag'}]}
        assert response.status_int == 200



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import operator
import textwrap

import docutils.core
from pecan import expose
from pecan.hooks import HookController
import routes

from fairy_slipper import hooks
from fairy_slipper.rest import JSONWriter
from paste.deploy import util as paste_util

LOG = logging.getLogger(__name__)


class VersionAPIController(object):

    def __init__(self, versions):
        self.versions = versions

    @expose()
    def _lookup(self, id_, *remainder):
        LOG.error(id_)
        if id_ in self.versions:
            return DocSpecController(id_, self.versions[id_]), remainder

    @expose(generic=True, template='json')
    def index(self):
        return self.versions.keys()


class DocSpecController(HookController):

    __hooks__ = [hooks.CORSHook()]

    def __init__(self, version, router):
        # TODO(RS) this had to be hardcoded, to match the murano
        # factory method.  Perhaps there is a better way to get it to
        # work using the factory?
        self.version = version
        self.api = paste_util.lookup_object(router)(routes.Mapper())
        super(DocSpecController, self).__init__()

    @expose(generic=True, template='json')
    def index(self):
        routes = {}
        for route in self.api.map.matchlist:
            if 'controller' not in route.defaults:
                continue

            key = (id(route.defaults['controller']),
                   route.defaults['action'])

            controller = route.defaults['controller'].controller
            action = route.defaults['action']

            if not hasattr(controller, action):
                continue

            if route.routepath.endswith('.:(format)'):
                continue

            if key not in routes:
                routes[key] = {'routepath': [],
                               'req': []}

            routes[key]['routepath'] = '/' + self.version + route.routepath
            routes[key]['req'] = route.reqs
            routes[key]['action'] = action
            routes[key]['conditions'] = route.conditions
            doc = getattr(controller, action).__doc__
            if doc:
                json = docutils.core.publish_parts(
                    textwrap.dedent(doc),
                    writer=JSONWriter())
                routes[key].update(json)
            routes[key]['classpath'] = '.'.join(
                [controller.__class__.__module__,
                 controller.__class__.__name__]
            ) + ':' + getattr(controller, action).__name__

        routes = sorted(routes.values(),
                        key=operator.itemgetter('classpath'))
        return routes



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import unicode_literals

import json
import logging
import os
from os import path
import re
import textwrap
import xml.sax

import prettytable

log = logging.getLogger(__name__)


SECTIONS = {'API_Versions': 'api-versions',
            'Database_Instances': 'database-instances',
            'admin-tenants': 'admin-tenants',
            'admin-tokens': 'admin-tokens',
            'admin-users': 'admin-users',
            'admin-versions': 'admin-versions',
            'alarms': 'alarms',
            'build-info': 'build-info',
            'capabilities': 'capabilities',
            'compute_extensions': 'extensions',
            'compute_flavors': 'flavors',
            'compute_image_metadata': 'image-metadata',
            'compute_images': 'images',
            'compute_limits': 'limits',
            'compute_server-actions': 'server-actions',
            'compute_server-addresses': 'server-addresses',
            'compute_server_metadata': 'server-metadata',
            'compute_servers': 'servers',
            'compute_versions': 'versions',
            'credentials-v3': 'credentials',
            'database-instance-actions': 'database-instance-actions',
            'databases': 'databases',
            'datastores': 'datastores',
            'diagnostics-v2.1': 'diagnostics',
            'domains-v3': 'domains',
            'endpoints-v3': 'endpoints',
            'ext-backups-v2': 'ext-backups',
            'extraroute-ext': 'extraroute',
            'flavor-extra-specs-v2.1': 'flavor-extra-specs',
            'flavors': 'flavors',
            'general-info': 'general-info',
            'groups-v3': 'groups',
            'heat-versions': 'heat-versions',
            'identity-auth-v2': 'identity-auth',
            'identity-v2-versions': 'versions',
            'identity_v3_OS-INHERIT-ext': 'inherit',
            'identity_v3_OS-KDS-ext': 'kds',
            'identity_v3_OS-OAUTH1-ext': 'oauth1',
            'identity_v3_OS-TRUST-ext': 'trust',
            'image-data-v2': 'image-data',
            'image-schemas-v2': 'image-schemas',
            'image-tags-v2': 'image-tags',
            'images-v1': 'images',
            'images-v2': 'images',
            'keypairs-v2.1': 'keypairs',
            'layer3': 'layer3',
            'lbaas-v1.0': 'lbaas',
            'lbaas-v2.0': 'lbaas',
            'limits-v2.1': 'limits',
            'members-v1': 'members',
            'members-v2': 'members',
            'metering-ext': 'metering',
            'meters': 'meters',
            'network_multi_provider-ext': 'network-multi-provider',
            'network_provider-ext': 'network-provider',
            'network_vlan_transparency-ext': 'network-vlan-transparency',
            'networks': 'networks',
            'neutron-versions-v2': 'versions',
            'neutron_extensions': 'extensions',
            'os-admin-actions': 'admin-actions',
            'os-admin-actions-v2.1': 'admin-actions',
            'os-admin-password-v2.1': 'admin-password',
            'os-agents': 'agents',
            'os-agents-v2.1': 'agents',
            'os-aggregates': 'aggregates',
            'os-aggregates-v2.1': 'aggregates',
            'os-availability-zone': 'availability-zone',
            'os-availability-zone-v2.1': 'availability-zone',
            'os-baremetal-ext-status': 'baremetal-ext-status',
            'os-block-device-mapping-v2-boot': 'block-device-mapping',
            'os-cells-v2.1': 'cells',
            'os-certificates': 'certificates',
            'os-certificates-v2.1': 'certificates',
            'os-cloudpipe': 'cloudpipe',
            'os-config-drive-v2.1': 'config-drive',
            'os-console-output': 'console-output',
            'os-console-output-v2.1': 'console-output',
            'os-consoles': 'consoles',
            'os-coverage': 'coverage',
            'os-create-backup-v2.1': 'create-backup',
            'os-createserverext': 'createserverext',
            'os-deferred-delete': 'deferred-delete',
            'os-deferred-delete-v2.1': 'deferred-delete',
            'os-diagnostics': 'diagnostics',
            'os-disk-config': 'disk-config',
            'os-evacuate-v2.1': 'evacuate',
            'os-ext-az': 'ext-az',
            'os-ext-img-size': 'ext-img-size',
            'os-ext-ips': 'ext-ips',
            'os-extended-availability-zone-v2.1': 'extended-availability-zone',
            'os-extended-networks': 'extended-networks',
            'os-extended-server-attributes': 'extended-server-attributes',
            'os-extended-server-attributes-v2.1': 'extended-server-attributes',
            'os-extended-status': 'extended-status',
            'os-extended-status-v2.1': 'extended-status',
            'os-fixed-ips': 'fixed-ips',
            'os-flavor-access': 'flavor-access',
            'os-flavor-access-v2.1': 'flavor-access',
            'os-flavor-extra-specs': 'flavor-extra-specs',
            'os-flavor-manage-v2.1': 'flavor-manage',
            'os-flavor-rxtx': 'flavor-rxtx',
            'os-flavor-rxtx-v2.1': 'flavor-rxtx',
            'os-flavor-swap': 'flavor-swap',
            'os-flavorextradata': 'flavorextradata',
            'os-flavormanage': 'flavormanage',
            'os-flavors-v2.1': 'flavors',
            'os-floating-ip-dns': 'floating-ip-dns',
            'os-floating-ip-dns-v2.1': 'floating-ip-dns',
            'os-floating-ip-pools': 'floating-ip-pools',
            'os-floating-ip-pools-v2.1': 'floating-ip-pools',
            'os-floating-ips': 'floating-ips',
            'os-floating-ips-bulk': 'floating-ips-bulk',
            'os-floating-ips-bulk-v2.1': 'floating-ips-bulk',
            'os-floating-ips-v2.1': 'floating-ips',
            'os-flv-disabled': 'flv-disabled',
            'os-hosts': 'hosts',
            'os-hosts-v2.1': 'hosts',
            'os-hypervisor-status': 'hypervisor-status',
            'os-hypervisors': 'hypervisors',
            'os-hypervisors-v2.1': 'hypervisors',
            'os-instance-actions': 'instance-actions',
            'os-instance-actions-v2.1': 'instance-actions',
            'os-instance-usage-audit-log-v2.1': 'instance-usage-audit-log',
            'os-interface': 'interface',
            'os-keypairs': 'keypairs',
            'os-ksadm-admin-ext': 'ksadm-admin',
            'os-kscatalog-ext': 'kscatalog',
            'os-ksec2-admin-ext': 'ksec2-admin',
            'os-kss3-admin-ext': 'kss3-admin',
            'os-ksvalidate-ext': 'ksvalidate',
            'os-limits-v2': 'limits',
            'os-metadef-namespace-v2': 'metadef-namespace',
            'os-metadef-object-v2': 'metadef-object',
            'os-metadef-property-v2.wadl': 'metadef-property',
            'os-metadef-resourcetype-v2': 'metadef-resourcetype',
            'os-metadef-schemas-v2': 'metadef-schemas',
            'os-metadef-tag-v2': 'metadef-tag',
            'os-migrations': 'migrations',
            'os-migrations-v2.1': 'migrations',
            'os-multi-server-create': 'multi-server-create',
            'os-multinic-v2.1': 'multinic',
            'os-multiple-create-v2.1': 'multiple-create',
            'os-networks': 'networks',
            'os-networks-v2.1': 'networks',
            'os-qos-v2-qos-specs': 'qos-v2-qos-specs',
            'os-quota-class-sets': 'quota-class-sets',
            'os-quota-class-sets-v2.1': 'quota-class-sets',
            'os-quota-sets': 'quota-sets',
            'os-quota-sets-v2': 'quota-sets',
            'os-quota-sets-v2.1': 'quota-sets',
            'os-remote-consoles-v2.1': 'remote-consoles',
            'os-rescue': 'rescue',
            'os-scheduler-hints': 'scheduler-hints',
            'os-security-group-default-rules': 'security-group-default-rules',
            'os-security-groups': 'security-groups',
            'os-security-groups-v2.1': 'security-groups',
            'os-server-OS-EXT-IPS-MAC': 'server-ext-ips-mac',
            'os-server-actions-v2.1': 'server-actions',
            'os-server-groups': 'server-groups',
            'os-server-groups-v2.1': 'server-groups',
            'os-server-password': 'server-password',
            'os-server-password-v2.1': 'server-password',
            'os-server-shelve': 'server-shelve',
            'os-server-start-stop': 'server-start-stop',
            'os-server-usage-v2.1': 'server-usage',
            'os-services': 'services',
            'os-services-v2.1': 'services',
            'os-shelve-v2.1': 'shelve',
            'os-simple-tenant-usage': 'simple-tenant-usage',
            'os-tenant-networks-v2.1': 'tenant-networks',
            'os-used-limits': 'used-limits',
            'os-used-limits-for-admins': 'used-limits-for-admins',
            'os-virtual-interfaces': 'virtual-interfaces',
            'os-virtual-interfaces-v2.1': 'virtual-interfaces',
            'os-volume': 'volume',
            'os-volume-manage-v2': 'volume-manage',
            'os-volume-type-access-v2': 'volume-type-access',
            'os-volume_attachments': 'volume-attachments',
            'policies-v3': 'policies',
            'port_binding-ext': 'port-binding',
            'ports': 'ports',
            'projects-v3': 'projects',
            'quotas-ext': 'quotas',
            'resources': 'resources',
            'roles-v3': 'roles',
            'samples': 'samples',
            'security_groups': 'security-groups',
            'server-ips-v2.1': 'server-ips',
            'service-catalog-v3': 'service-catalog',
            'service-status': 'service-status',
            'shared_images_v1': 'shared-images',
            'software-config': 'software-config',
            'stack-actions': 'stack-actions',
            'stack-events': 'stack-events',
            'stack-resources': 'stack-resources',
            'stack-templates': 'stack-templates',
            'stacks': 'stacks',
            'storage_account_services': 'storage-account-services',
            'storage_container_services': 'storage-container-services',
            'storage_object_services': 'storage-object-services',
            'subnets': 'subnets',
            'tokens-v3': 'tokens',
            'user_management': 'user-management',
            'users-v3': 'users',
            'v1.1clusters': 'clusters',
            'v1.1clustertemplate': 'clustertemplate',
            'v1.1datasources': 'datasources',
            'v1.1event-log': 'event-log',
            'v1.1imageregistry': 'imageregistry',
            'v1.1jobbinaries': 'jobbinaries',
            'v1.1jobbinary-internals': 'jobbinary-internals',
            'v1.1jobexecutions': 'job-executions',
            'v1.1jobs': 'jobs',
            'v1.1jobtypes': 'job-types',
            'v1.1nodegrouptemplate': 'node-group-template',
            'v1.1plugins': 'plugins',
            'v2.1os-fping': 'fping',
            'versions-identity-v3': 'versions-identity',
            'versions-images-v2': 'versions-images',
            'versions-v1': 'versions',
            'versions-v2.1': 'versions',
            'volume-api-v1-snapshots': 'snapshots',
            'volume-api-v1-types': 'types',
            'volume-api-v1-versions': 'versions',
            'volume-api-v1-volumes': 'volumes',
            'volume-api-v2-extensions': 'extensions',
            'volume-api-v2-snapshots': 'snapshots',
            'volume-api-v2-types': 'types',
            'volume-api-v2-versions': 'versions',
            'volume-api-v2-volumes': 'volumes',
            'vpnaas-v2.0': 'vpnaas'}


VERSION_RE = re.compile('v[0-9\.]+')
WHITESPACE_RE = re.compile('[\s]+', re.MULTILINE)
TITLE_RE = re.compile(
    '(.*) API (v([\d.]+) )?(\S*)[ ]*\((SUPPORTED|CURRENT|DEPRECATED|\
EXPERIMENTAL)\)')
CAPTION_RE = re.compile('[*`]*')
MARKUP_RE = re.compile('[.,:;)]+')


class TableMixin(object):
    def visit_table(self, attrs):
        self.__table = prettytable.PrettyTable(hrules=prettytable.ALL)
        self.__table.header = False

    def depart_table(self):
        self.content.append('\n\n')
        self.content.append(str(self.__table))
        self.content.append('\n\n')

    def visit_caption(self, attrs):
        self.content_stack.append([])

    def depart_caption(self):
        content = ''.join(self.content_stack.pop()).strip()
        content = CAPTION_RE.sub('', content)
        content = WHITESPACE_RE.sub(' ', content)
        content = '**' + content + '**'
        self.content.append(content)

    def visit_th(self, attrs):
        self.__table.header = True

    def depart_th(self):
        heading = self.content.pop().strip()
        self.__table.field_names.append(heading)
        self.__table.align[heading] = 'l'
        self.__table.valign[heading] = 't'
        self.__table.max_width[heading] = 80

    def visit_tr(self, attrs):
        self.__row = []

    def visit_td(self, attrs):
        self.content_stack.append([])

    def depart_td(self):
        self.__row.append(''.join(self.content_stack.pop()).strip())

    def depart_tr(self):
        if self.__row:
            columns = len(self.__table.field_names)
            self.__row.extend(['' for n in range(columns - len(self.__row))])
            self.__table.add_row(self.__row)


class APIChapterContentHandler(xml.sax.ContentHandler, TableMixin):

    EMPHASIS = {
        'bold': '**',
        'italic': '*'
    }

    def __init__(self, filename, api_parser):
        self.filename = filename
        self.api_parser = api_parser

    def startDocument(self):
        super(APIChapterContentHandler, self).startDocument()
        self.tags = {}
        self.current_tag = None

        # general state
        self.tag_stack = []
        self.attr_stack = []
        self.content_stack = [[]]
        self.current_emphasis = None
        self.nesting = 0
        self.no_space = False
        self.fill_width = 67
        self.wrapper = textwrap.TextWrapper(width=self.fill_width)
        self.inline_markup_stack = []
        self.base_indent = ' '
        self.hyperlink_end = False
        self.markup_end = False

    @property
    def content(self):
        return self.content_stack[-1]

    def search_stack_for(self, tag_name):
        for tag, attrs in zip(reversed(self.tag_stack),
                              reversed(self.attr_stack)):
            if tag == tag_name:
                return attrs

    def on_top_tag_stack(self, *args):
        return self.tag_stack[-len(args):] == list(args)

    def startElement(self, name, _attrs):
        attrs = dict(_attrs)

        self.tag_stack.append(name)
        self.attr_stack.append(attrs)

        if self.on_top_tag_stack('chapter', 'section', 'title'):
            self.content_stack.append([])

        if self.on_top_tag_stack('chapter', 'title'):
            self.content_stack.append([])

        if self.on_top_tag_stack('chapter', 'section'):
            self.content_stack.append([])
            id = attrs['xml:id']
            id = SECTIONS.get(id, id)
            self.current_tag = {'name': id}
            self.api_parser.tags.append(self.current_tag)

        if name == 'wadl:resource':
            filename, resource_id = attrs['href'].split("#")
            dir = path.dirname(self.filename)
            filepath = path.abspath(path.join(dir, filename))
            tag_name = '%s#%s' % (filepath, resource_id)
            self.api_parser.resource_tags[tag_name] = self.current_tag['name']

        if name == 'wadl:resources':
            if 'href' in attrs:
                dir = path.dirname(self.filename)
                filepath = path.abspath(path.join(dir, attrs['href']))
                self.api_parser.file_tags[filepath] = self.current_tag['name']

        if self.on_top_tag_stack('wadl:resource', 'wadl:method'):
            resource = self.search_stack_for('wadl:resource')
            dir = path.dirname(self.filename)
            filename = resource['href'].split("#")[0]
            filepath = path.abspath(path.join(dir, filename))
            method_path = filepath + attrs['href']
            self.api_parser.method_tags[method_path] = self.current_tag['name']

        fn = getattr(self, 'visit_%s' % name, None)
        if fn:
            fn(dict(_attrs))

    def endElement(self, name):
        content = ''.join(self.content)

        if self.on_top_tag_stack('chapter', 'section', 'title'):
            self.current_tag['description'] = content.strip()
            self.content_stack.pop()

        if self.on_top_tag_stack('chapter', 'title'):
            title = content.strip()
            match = TITLE_RE.match(title)
            if match:
                title, version, version_num, ext, state = match.groups()
            else:
                raise Exception("Title %s doesn't match RE" % title)
            self.api_parser.title = ('%s %s' % (title, ext)).strip()
            self.content_stack.pop()

        if self.on_top_tag_stack('chapter', 'section'):
            self.current_tag['summary'] = content.strip()
            self.content_stack.pop()

        self.tag_stack.pop()
        self.attr_stack.pop()

        fn = getattr(self, 'depart_%s' % name, None)
        if fn:
            fn()

    def characters(self, content):
        if not content:
            return
        # Fold up any white space into a single char
        if not self.on_top_tag_stack('programlisting'):
            content = WHITESPACE_RE.sub(' ', content)

        if content == ' ':
            return
        if content[0] == '\n':
            return
        if self.content:
            if self.content[-1].endswith('\n'):
                content = ' ' * self.nesting + content.strip()
            elif self.content[-1].endswith(' '):
                content = content.strip()
            elif (self.on_top_tag_stack('programlisting')):
                content = '\n' + ' ' * self.nesting + content
            elif self.no_space:
                content = '' + content.strip()
            elif self.hyperlink_end:
                self.hyperlink_end = False
                if content == '.' or content == ':':
                    pass
                else:
                    content = ' ' + content.strip()
            elif self.markup_end:
                self.markup_end = False
                if MARKUP_RE.match(content):
                    pass
                else:
                    content = ' ' + content.strip()
            else:
                content = ' ' + content.strip()

        if self.no_space is True:
            self.inline_markup_stack.append(content)
        else:
            self.content.append(content)

    def visit_listitem(self, attrs):
        self.nesting = len([tag for tag in self.tag_stack
                            if tag == 'listitem']) - 1
        if self.nesting > 0:
            prev_nesting = self.nesting - 1
            self.base_indent = ' ' * prev_nesting + '  '
        else:
            self.base_indent = ' '

        self.content_stack.append([self.base_indent * self.nesting + '-'])
        self.wrapper = textwrap.TextWrapper(
            width=self.fill_width,
            initial_indent=' ',
            subsequent_indent=self.base_indent * self.nesting + '  ',)

    def depart_listitem(self):
        content = self.content_stack.pop()
        self.content.append(''.join(content))
        if self.content[-1].endswith('\n\n'):
            pass
        else:
            self.content.append('\n')

        self.nesting = len([tag for tag in self.tag_stack
                            if tag == 'listitem']) - 1
        if self.nesting > 0:
            prev_nesting = self.nesting - 1
            self.base_indent = ' ' * prev_nesting + '  '
        else:
            self.base_indent = ' '

    def depart_itemizedlist(self):
        if self.search_stack_for('itemizedlist') is None:
            self.wrapper = textwrap.TextWrapper(width=self.fill_width)
        else:
            self.wrapper = textwrap.TextWrapper(
                width=self.fill_width,
                initial_indent=self.base_indent * self.nesting + '  ',
                subsequent_indent=self.base_indent * self.nesting + '  ',)

    def depart_orderedlist(self):
        if self.search_stack_for('itemizedlist') is None:
            self.wrapper = textwrap.TextWrapper(width=self.fill_width)

    def visit_para(self, attrs):
        self.content_stack.append([''])
        if self.search_stack_for('itemizedlist') is not None:
            return
        if self.content:
            if self.content[-1].endswith('\n\n'):
                pass
            elif self.content[-1].endswith('\n'):
                self.content.append('\n')

    def depart_para(self):
        content = ''.join(self.content_stack.pop()).strip()
        wrapped = self.wrapper.wrap(content)
        self.content.append('\n'.join(wrapped))
        self.content.append('\n\n')

        if self.search_stack_for('itemizedlist') is None:
            pass
        else:
            self.wrapper = textwrap.TextWrapper(
                width=self.fill_width,
                initial_indent=self.base_indent * self.nesting + '  ',
                subsequent_indent=self.base_indent * self.nesting + '  ',)

    def visit_code(self, attrs):
        self.no_space = True

    def depart_code(self):
        content = ' ``'
        if self.content:
            if self.content[-1].endswith('(') or \
               self.content[-1].endswith(' '):
                content = '``'
        content += ' '.join(self.inline_markup_stack[0:None])
        content += '``'
        self.content.append(content)
        self.inline_markup_stack[:] = []
        self.no_space = False
        self.markup_end = True

    def visit_emphasis(self, attrs):
        # Bold is the default emphasis
        self.current_emphasis = attrs.get('role', 'bold')
        self.no_space = True

    def depart_emphasis(self):
        content = ' ' + self.EMPHASIS[self.current_emphasis]
        if self.content:
            if self.content[-1].endswith('(') or \
               self.content[-1].endswith(' '):
                content = '' + self.EMPHASIS[self.current_emphasis]
        content += ' '.join(self.inline_markup_stack[0:None])
        content += self.EMPHASIS[self.current_emphasis]
        self.content.append(content)
        self.inline_markup_stack[:] = []
        self.no_space = False
        self.current_emphasis = None
        self.markup_end = True

    def visit_programlisting(self, attrs):
        if not attrs:
            self.content.append('::\n\n')
        else:
            self.content.append('.. code-block:: %s\n\n' % attrs['language'])
        self.nesting = 3

    def depart_programlisting(self):
        self.nesting = 0  # no indent for blank lines
        self.content.append('\n\n')

    def visit_link(self, attrs):
        if attrs:
            self.inline_markup_stack.append(attrs['xlink:href'])
            self.no_space = True

    def depart_link(self):
        content = ' `'
        # anonymous link
        if len(self.inline_markup_stack) is 1:
            content += ('<%s>`__' % self.inline_markup_stack[0])
        else:
            content += ' '.join(self.inline_markup_stack[1:None])
            content += (' <%s>`_' % self.inline_markup_stack[0])

        self.content.append(content)
        self.inline_markup_stack[:] = []
        self.no_space = False
        self.hyperlink_end = True


class APIRefContentHandler(xml.sax.ContentHandler):

    def __init__(self, filename):
        self.filename = filename

    def startDocument(self):
        self.tags = []
        self.current_tag = None
        self.method_tags = {}
        self.resource_tags = {}
        self.file_tags = {}

        # general state
        self.tag_stack = []
        self.attr_stack = []
        self.content = None

    def search_stack_for(self, tag_name):
        for tag, attrs in zip(reversed(self.tag_stack),
                              reversed(self.attr_stack)):
            if tag == tag_name:
                return attrs

    def on_top_tag_stack(self, *args):
        return self.tag_stack[-len(args):] == list(args)

    def startElement(self, name, _attrs):
        attrs = dict(_attrs)
        self.tag_stack.append(name)
        self.attr_stack.append(attrs)
        self.content = []
        if self.on_top_tag_stack('book'):
            id = attrs['xml:id']
            extensions = False
            if id.endswith('-ext'):
                extensions = True
                id = id.rsplit('-', 1)[0]
            service, version = id.rsplit('-', 1)
            if service.startswith('api.openstack.org-'):
                service = service.split('-', 1)[1]
            if extensions:
                service = service + '-extensions'
            assert VERSION_RE.match(version)
            self.service = service
            self.version = version
        if self.on_top_tag_stack('book', 'xi:include'):
            filename = attrs['href']
            dir = path.dirname(self.filename)
            filepath = path.join(dir, filename)
            ch = APIChapterContentHandler(filepath, self)
            xml.sax.parse(filepath, ch)

    def endElement(self, name):
        self.tag_stack.pop()
        self.attr_stack.pop()

    def characters(self, content):
        content = content.strip()
        if content:
            self.content.append(content)


def main1(source_file, output_dir):
    log.info('Parsing %s' % source_file)
    ch = APIRefContentHandler(source_file)
    xml.sax.parse(source_file, ch)
    os.chdir(output_dir)
    output = {
        'title': ch.title,
        'service': ch.service,
        'version': ch.version,
        'tags': ch.tags,
        'method_tags': ch.method_tags,
        'file_tags': ch.file_tags,
        'resource_tags': ch.resource_tags,
    }
    pathname = 'api-ref-%s-%s.json' % (ch.service,
                                       ch.version)
    with open(pathname, 'w') as out_file:
        json.dump(output, out_file, indent=2, sort_keys=True)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-v', '--verbose', action='count', default=0,
        help="Increase verbosity (specify multiple times for more)")
    parser.add_argument(
        '-o', '--output-dir', action='store',
        help="The directory to output the JSON files too.")
    parser.add_argument(
        'filename',
        help="File to convert")

    args = parser.parse_args()

    log_level = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG

    logging.basicConfig(
        level=log_level,
        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    filename = path.abspath(args.filename)

    main1(filename, output_dir=args.output_dir)



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import logging

import docutils.core
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
import docutils.utils
from docutils import writers
import six

logger = logging.getLogger(__name__)


MIME_MAP = {
    'json': 'application/json',
    'txt': 'text/plain',
    'xml': 'application/xml',
}

STATUS_CODE_MAP = {
    '200': 'Success',
    '201': 'Created',
    '202': 'Accepted',
    '203': 'Non-Authoritative Information',
    '204': 'No Content',
    '205': 'Reset Content',
    '206': 'Partial Content',
    '300': 'Multiple Choices',
    '301': 'Moved Permanently',
    '302': 'Found',
    '303': 'See Other',
    '304': 'Not Modified',
    '400': 'Bad Request',
    '401': 'Unauthorized',
    '403': 'Forbidden',
    '404': 'Not Found',
    '405': 'Method Not Allowed',
    '409': 'Conflict',
    '410': 'Gone',
    '413': 'Request Entity Too Large',
    '415': 'Unsupported Media Type',
    '503': 'Service Unavailable',
}


def search_node_parents(node, node_name):
    parent = node
    while parent.parent:
        if parent.tagname == node_name:
            return node
        parent = parent.parent
    if parent.tagname == node_name:
        return node


class JSONTranslator(nodes.GenericNodeVisitor):
    def __init__(self, document):
        nodes.NodeVisitor.__init__(self, document)
        self.output = {
            'tags': [],
            'paths': {}
        }
        self.node_stack = []
        self.node_stack.append(self.output)
        self.current_node_name = None
        self.bullet_stack = []
        self.table_stack = []
        self.text = ''
        self.col_num = 0
        self.first_row = 0
        self.hyperlink_name = ''
        self.refuri = ''
        self.listitem = False
        self.lit_block = False
        self.list_indent = 0
        self.text_res_desc = ''

    def search_stack_for(self, tag_name):
        for node in self.node_stack:
            # Skip any list elements, this is a hack, but it' works
            # for now.
            if isinstance(node, (list, ) + six.string_types):
                continue
            if tag_name in node.keys():
                return node

    def visit_document(self, node):
        # Disable both the document visit and depart
        pass

    def depart_document(self, node):
        pass

    def default_visit(self, node):
        """Default node visit method."""
        self.current_node_name = node.__class__.__name__
        if hasattr(node, 'children') and node.children:
            new_node = {}
            self.node_stack[-1][self.current_node_name] = new_node
            self.node_stack.append(new_node)

    def default_departure(self, node):
        """Default node depart method."""
        if hasattr(node, 'children') and node.children:
            self.node_stack.pop()

    def visit_system_message(self, node):
        pass

    def depart_system_message(self, node):
        pass

    def visit_Text(self, node):
        if self.first_row is 0:
            if self.lit_block and len(self.bullet_stack) > 0:
                litblock = node.astext().split('\n')
                litblock = '\n        '.join(litblock)
                self.text += litblock
            else:
                self.text += node.astext()

    def depart_Text(self, node):
        pass

    def visit_emphasis(self, node):
        if self.first_row > 0:
            inlinetxt = self.table_stack.pop()
            para = inlinetxt.partition(node.astext())
            new_para = ''
            new_para += para[0] + '_' + para[1] + '_' + para[2]
            self.table_stack.append(new_para)
        else:
            self.text += '_'

    def depart_emphasis(self, node):
        if self.first_row is 0:
            self.text += '_'

    def visit_literal(self, node):
        if self.first_row > 0:
            inlinetxt = self.table_stack.pop()
            para = inlinetxt.partition(node.astext())
            new_para = ''
            new_para += para[0] + '`' + para[1] + '`' + para[2]
            self.table_stack.append(new_para)
        else:
            self.text += '`'

    def depart_literal(self, node):
        if self.first_row is 0:
            self.text += '`'

    def visit_strong(self, node):
        if self.first_row > 0:
            inlinetxt = self.table_stack.pop()
            para = inlinetxt.partition(node.astext())
            new_para = ''
            new_para += para[0] + '**' + para[1] + '**' + para[2]
            self.table_stack.append(new_para)
        else:
            self.text += '**'

    def depart_strong(self, node):
        if self.first_row is 0:
            self.text += '**'

    def visit_literal_block(self, node):
        if len(self.bullet_stack) > 0:
            self.text += '\n        '
        else:
            self.text += '```\n'
        self.lit_block = True

    def depart_literal_block(self, node):
        if len(self.bullet_stack) > 0:
            self.text += '\n'
        else:
            self.text += '\n```\n'
        self.lit_block = False

    def visit_bullet_list(self, node):
        if self.first_row > 0:
            self.text += """<ul>"""
        else:
            self.bullet_stack.append('*')

    def depart_bullet_list(self, node):
        if self.first_row > 0:
            self.text += """</ul>"""
        else:
            self.bullet_stack.pop()
            self.list_indent = len(self.bullet_stack) - 1
            if len(self.bullet_stack) is 0:
                self.text += '\n'

    def visit_list_item(self, node):
        if self.first_row > 0:
            self.text += """<li>"""
        else:
            self.list_indent = len(self.bullet_stack) - 1
            item = '\n%s%s ' % ('  ' * self.list_indent,
                                self.bullet_stack[-1])
            self.text += item
            self.listitem = True

    def depart_list_item(self, node):
        if self.first_row > 0:
            self.text += """</li>"""
        else:
            self.listitem = False
            self.list_indent = 0

    def visit_title(self, node):
        self.current_node_name = node.__class__.__name__
        if self.current_node_name not in self.node_stack[-1]:
            new_node = []
            self.node_stack[-1][self.current_node_name] = new_node
            self.node_stack.append(new_node)

    def depart_title(self, node):
        self.node_stack.pop()

    def visit_paragraph(self, node):
        if self.first_row > 0:
            self.table_stack.append(node.astext())
        else:
            # listitem text
            if self.listitem is True:
                pass
            else:
                # another para in listitem
                if len(self.bullet_stack) > 0:
                    if self.lit_block:
                        self.text += '\n' + '        '
                    else:
                        self.text += '\n' + '  ' * self.list_indent + ' '

    def depart_paragraph(self, node):
        if self.first_row is 0:
            if self.listitem:
                self.text += '\n'
                self.listitem = False
            else:
                if len(self.bullet_stack) > 0:
                    self.text += "\n"
                else:
                    # default paragraph
                    self.text += "\n\n"
        else:
            if self.first_row > 0:
                para = self.table_stack.pop()
                para = para.strip('\n')
                plist = para.split('\n')

                # multi-line text in single column
                if len(plist) > 0:
                    self.text += """<br>""".join(plist)
                else:
                    self.text += para

    def visit_line_block(self, node):
        if isinstance(self.node_stack[-1], list):
            return

        self.current_node_name = node.__class__.__name__
        if self.current_node_name not in self.node_stack[-1]:
            new_node = []
            self.node_stack[-1][self.current_node_name] = new_node
            self.node_stack.append(new_node)
        else:
            self.node_stack.append(self.node_stack[-1][self.current_node_name])

    def depart_line_block(self, node):
        if isinstance(self.node_stack[-1], list):
            self.node_stack.pop()

    def visit_table(self, node):
        self.col_num = 0

    def depart_table(self, node):
        self.text += "\n"

    def visit_tbody(self, node):
        pass

    def depart_tbody(self, node):
        self.text += "\n"
        self.first_row = 0
        self.col_num = 0

    def visit_thead(self, node):
        pass

    def depart_thead(self, node):
        pass

    def visit_tgroup(self, node):
        pass

    def depart_tgroup(self, node):
        pass

    def visit_colspec(self, node):
        pass

    def depart_colspec(self, node):
        pass

    def visit_row(self, node):
        if self.first_row is 1 and self.col_num > 0:
            row_separator = [' --- '] * self.col_num
            self.text += "|"
            sep_row = "|".join(row_separator)
            self.text += sep_row
            self.text += "|"
            self.text += "\n"

        self.text += "|"
        self.first_row += 1

    def depart_row(self, node):
        self.text += "\n"

    def visit_entry(self, node):
        self.text += " "

    def depart_entry(self, node):
        self.text += " |"
        self.col_num += 1

    def visit_definition(self, node):
        pass

    def depart_definition(self, node):
        pass

    def visit_definition_list(self, node):
        pass

    def depart_definition_list(self, node):
        pass

    def visit_definition_list_item(self, node):
        pass

    def depart_definition_list_item(self, node):
        pass

    def visit_term(self, node):
        self.text += "    "
        if self.first_row is 0:
            self.text += node.astext()
        else:
            self.table_stack.append(node.astext())

    def depart_term(self, node):
        if self.first_row > 0:
            self.text += self.table_stack.pop()
        self.text += """<br>"""

    def visit_reference(self, node):
        self.hyperlink_name = node.attributes['name']
        self.refuri = node.attributes['refuri']
        self.text += '['

    def depart_reference(self, node):
        if self.hyperlink_name:
            self.text += ']'
            self.text += '(' + self.refuri + ')'
        else:
            self.text += '[' + self.refuri + ']'

        self.hyperlink_name = ''
        self.refuri = ''

    def visit_resource(self, node):
        self.text = ''
        if 'paths' not in self.node_stack[-1]:
            self.node_stack[-1]['paths'] = {}
        self.node_stack.append(self.node_stack[-1]['paths'])

    def depart_resource(self, node):
        self.node_stack[-1]['description'] = self.text
        # XXX This is a massive hack, this is here because the visit
        # resource url functions don't pop the stack.
        self.node_stack.pop()
        self.node_stack.pop()

    def visit_resource_url(self, node):
        url_path = node.astext()
        node.clear()
        if url_path not in self.node_stack[-1]:
            self.node_stack[-1][url_path] = []
        new_node = {'responses': {},
                    'parameters': [],
                    'description': '',
                    'produces': [],
                    'consumes': [],
                    'tags': []}
        self.node_stack[-1][url_path].append(new_node)
        self.node_stack.append(new_node)

    def depart_resource_url(self, node):
        pass

    def visit_resource_summary(self, node):
        summary = node.astext()
        self.node_stack[-1]['summary'] = summary
        node.clear()

    def depart_resource_summary(self, node):
        pass

    def visit_resource_title(self, node):
        title = node.astext()
        # Should probably be x-title
        self.node_stack[-1]['title'] = title
        node.clear()

    def depart_resource_title(self, node):
        pass

    def visit_resource_method(self, node):
        method = node.astext()
        self.node_stack[-1]['method'] = method
        node.clear()

    def depart_resource_method(self, node):
        pass

    def visit_field_list(self, node):
        pass

    def depart_field_list(self, node):
        pass

    def visit_field(self, node):
        name = node.attributes['names'][0]
        resource = self.node_stack[-1]
        new_response = {'description': ''}
        # TODO(arrsim) this name matching ignores all the other
        # possible names that the fields could have.
        if name == 'statuscode':
            responses = resource['responses']
            status_code = node[0].astext()
            description = node[1].astext()
            if status_code not in responses:
                responses[status_code] = new_response
            if not description and status_code in STATUS_CODE_MAP:
                description = STATUS_CODE_MAP[status_code]
            responses[status_code]['description'] = description
            node.clear()
        elif name == 'responseexample':
            responses = resource['responses']
            status_code = node[0].astext()
            filepath = node[1].astext()
            if status_code not in responses:
                responses[status_code] = new_response
            ext = filepath.rsplit('.', 1)[1]
            mimetype = MIME_MAP[ext]
            if 'examples' not in responses[status_code]:
                responses[status_code]['examples'] = {}
            responses[status_code]['examples'][mimetype] = {'$ref': filepath}
            node.clear()
        elif name == 'requestexample':
            status_code = node[0].astext()
            filepath = node[1].astext()
            ext = filepath.rsplit('.', 1)[1]
            mimetype = MIME_MAP[ext]
            if 'examples' not in resource:
                resource['examples'] = {}
            resource['examples'][mimetype] = {'$ref': filepath}
            node.clear()
        elif name == 'requestschema':
            filepath = node[1].astext()
            resource['parameters'].append(
                {'name': 'body',
                 'in': 'body',
                 'required': True,
                 'schema': {'$ref': filepath}})
            node.clear()
        elif name == 'responseschema':
            responses = resource['responses']
            status_code = node[0].astext()
            filepath = node[1].astext()
            if status_code not in responses:
                responses[status_code] = new_response
            if 'schema' not in responses[status_code]:
                responses[status_code]['schema'] = {}
            responses[status_code]['schema'] = {'$ref': filepath}
            node.clear()
        elif name == 'parameter':
            param_name = node[0].astext()
            description = node[1].astext()
            resource['parameters'].append(
                {'name': param_name,
                 'description': description,
                 'in': 'path',
                 'type': 'string',
                 'required': True})
            node.clear()
        elif name == 'query':
            self.text_res_desc = self.text
            param_name = node[0].astext()
            self.text = ''
            description = ''
            resource['parameters'].append(
                {'name': param_name,
                 'description': description,
                 'in': 'query',
                 'type': 'string',
                 'required': False})
        elif name == 'reqheader':
            self.text_res_desc = self.text
            param_name = node[0].astext()
            self.text = ''
            description = ''
            resource['parameters'].append(
                {'name': param_name,
                 'description': description,
                 'in': 'header',
                 'type': 'string',
                 'required': False})
        elif name == 'tag':
            tag = node[1].astext()
            resource['tags'].append(tag)
            node.clear()
        elif name == 'accepts':
            mimetype = node[1].astext()
            resource['consumes'].append(mimetype)
            node.clear()
        elif name == 'produces':
            mimetype = node[1].astext()
            resource['produces'].append(mimetype)
            node.clear()
        else:
            node.clear()

    def depart_field(self, node):
        name = node.attributes['names'][0]
        resource = self.node_stack[-1]
        if name == 'query' or name == 'reqheader':
            param_name = node[0].astext()
            if self.text.startswith(param_name):
                resource['parameters'][-1]['description'] \
                    = self.text[len(param_name):]
            else:
                resource['parameters'][-1]['description'] = self.text
            self.text = self.text_res_desc

    def visit_field_name(self, node):
        self.node_stack[-1]['name'] = node.astext()

    def depart_field_name(self, node):
        pass

    def visit_field_body(self, node):
        self.node_stack[-1]['type'] = node.astext()

    def depart_field_body(self, node):
        pass

    def visit_field_type(self, node):
        self.node_stack[-1]['type'] = node.astext()

    def depart_field_type(self, node):
        pass

    def visit_swagger_tag(self, node):
        self.text = ''
        self.node_stack.append(self.node_stack[-1]['tags'])
        new_node = {'name': '',
                    'description': ''}
        self.node_stack[-1].append(new_node)
        self.node_stack.append(new_node)

    def depart_swagger_tag(self, node):
        self.node_stack[-1]['description'] = self.text
        self.node_stack.pop()
        self.node_stack.pop()

    def visit_swagger_tag_name(self, node):
        name = node.astext()
        node.clear()
        self.node_stack[-1]['name'] = name

    def depart_swagger_tag_name(self, node):
        pass

    def visit_swagger_tag_summary(self, node):
        summary = node.astext()
        node.clear()
        self.node_stack[-1]['summary'] = summary

    def depart_swagger_tag_summary(self, node):
        pass


class JSONWriter(writers.Writer):

    supported = ('json',)
    """Formats this writer supports."""

    settings_spec = (
        '"Docutils JSON" Writer Options',
        None,
        [])

    config_section = 'docutils_json writer'
    config_section_dependencies = ('writers',)

    output = None

    def __init__(self):
        writers.Writer.__init__(self)
        self.translator_class = JSONTranslator

    def translate(self):
        self.visitor = visitor = self.translator_class(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.output


class field_type(nodes.Part, nodes.TextElement):
    pass


class resource(nodes.Inline, nodes.TextElement):
    pass


class resource_url(nodes.Admonition, nodes.TextElement):
    pass


class resource_title(nodes.Admonition, nodes.TextElement):
    pass


class resource_summary(nodes.Admonition, nodes.TextElement):
    pass


class resource_method(nodes.Admonition, nodes.TextElement):
    pass


class Field(object):
    def __init__(self, name, names=(), label=None,
                 has_arg=True, rolename=None):
        self.name = name
        self.names = names
        self.label = label
        self.has_arg = has_arg
        self.rolename = rolename

    @classmethod
    def transform(cls, node):
        node.attributes['names'].append(node[0].astext())


class TypedField(Field):
    def __init__(self, name, names=(), label=None,
                 has_arg=True, rolename=None,
                 typerolename='', typenames=()):
        super(TypedField, self).__init__(
            name=name,
            names=names,
            label=label,
            has_arg=has_arg,
            rolename=rolename)
        self.typerolename = typerolename
        self.typenames = typenames

    @classmethod
    def transform(cls, node):
        split = node[0].rawsource.split(None, 2)
        type = None
        if len(split) == 3:
            name, type, value = split
        elif len(split) == 2:
            name, value = split
        else:
            raise Exception('Too Few arguments.')
        node.attributes['names'].append(name)
        if type:
            node.insert(1, field_type(type))
        node[0].replace_self(nodes.field_name(value, value))


class GroupedField(Field):

    @classmethod
    def transform(cls, node):
        name, value = node[0].rawsource.split(None, 1)
        node.attributes['names'].append(name)
        node[0].replace_self(nodes.field_name(value, value))


class Resource(Directive):

    method = None

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    final_argument_whitespace = True

    doc_field_types = [
        TypedField('parameter', label='Parameters',
                   names=('param', 'parameter', 'arg', 'argument'),
                   typerolename='obj', typenames=('paramtype', 'type')),
        TypedField('jsonparameter', label='JSON Parameters',
                   names=('jsonparameter', 'jsonparam', 'json'),
                   typerolename='obj',
                   typenames=('jsonparamtype', 'jsontype')),
        TypedField('requestjsonobject', label='Request JSON Object',
                   names=('reqjsonobj', 'reqjson', '<jsonobj', '<json'),
                   typerolename='obj', typenames=('reqjsonobj', '<jsonobj')),
        TypedField('requestjsonarray', label='Request JSON Array of Objects',
                   names=('reqjsonarr', '<jsonarr'),
                   typerolename='obj',
                   typenames=('reqjsonarrtype', '<jsonarrtype')),
        TypedField('responsejsonobject', label='Response JSON Object',
                   names=('resjsonobj', 'resjson', '>jsonobj', '>json'),
                   typerolename='obj', typenames=('resjsonobj', '>jsonobj')),
        TypedField('responsejsonarray', label='Response JSON Array of Objects',
                   names=('resjsonarr', '>jsonarr'),
                   typerolename='obj',
                   typenames=('resjsonarrtype', '>jsonarrtype')),
        TypedField('queryparameter', label='Query Parameters',
                   names=('queryparameter', 'queryparam', 'qparam', 'query'),
                   typerolename='obj',
                   typenames=('queryparamtype', 'querytype', 'qtype')),
        GroupedField('formparameter', label='Form Parameters',
                     names=('formparameter', 'formparam', 'fparam', 'form')),
        GroupedField('requestheader', label='Request Headers',
                     rolename='header',
                     names=('<header', 'reqheader', 'requestheader')),
        GroupedField('responseheader', label='Response Headers',
                     rolename='header',
                     names=('>header', 'resheader', 'responseheader')),
        GroupedField('statuscode', label='Status Codes',
                     rolename='statuscode',
                     names=('statuscode', 'status', 'code')),
        GroupedField('responseschema', label='Response Schema',
                     rolename='responseschema',
                     names=('reponse-schema', 'responseschema')),

        # Swagger Extensions
        GroupedField('responseexample', label='Response Example',
                     rolename='responseexample',
                     names=('swagger-response', 'responseexample')),
        Field('requestexample', label='Request Example',
              rolename='requestexample',
              names=('swagger-request', 'requestexample')),
        Field('requestschema', label='Request Schema',
              rolename='requestschema',
              names=('swagger-schema', 'requestschema')),
        Field('tag',
              label='Swagger Tag',
              rolename='tag',
              names=('swagger-tag', 'tag')),
        Field('accepts',
              label='Swagger Consumes',
              rolename='accepts',
              names=('swagger-accepts', 'accepts')),
        Field('produces',
              label='Swagger Consumes',
              rolename='produces',
              names=('swagger-produces', 'produces'))
    ]

    option_spec = {
        'title': lambda x: x,
        'synopsis': lambda x: x,
    }

    def transform_fields(self):
        return {name: f
                for f in self.doc_field_types
                for name in f.names}

    def run(self):
        node = resource()
        self.state.nested_parse(self.content, self.content_offset, node)
        fields = self.transform_fields()

        # This is the first line of the definition.
        url = self.arguments[0]
        node.insert(0, resource_url(url, url))

        if not node.children:
            return [node]

        if node[0].tagname == 'system_message':
            logger.error(node[0].astext())
            node.remove(node[0])

        # Method
        node.insert(1, resource_method(self.method, self.method))

        # Summary
        summary = self.options.get('synopsis', '')
        node.insert(1, resource_summary(summary, summary))
        title = self.options.get('title', '')
        node.insert(1, resource_title(title, title))

        # Generate field lists
        for child in node:
            if isinstance(child, nodes.field_list):
                for field in child:
                    name = field[0].rawsource.split(None, 1)[0]
                    fields[name].transform(field)
        return [node]


class HTTPGet(Resource):

    method = 'get'


class HTTPPost(Resource):

    method = 'post'


class HTTPPut(Resource):

    method = 'put'


class HTTPPatch(Resource):

    method = 'patch'


class HTTPOptions(Resource):

    method = 'options'


class HTTPHead(Resource):

    method = 'head'


class HTTPDelete(Resource):

    method = 'delete'


class HTTPCopy(Resource):

    method = 'copy'


directives.register_directive('http:get', HTTPGet)
directives.register_directive('http:post', HTTPPost)
directives.register_directive('http:put', HTTPPut)
directives.register_directive('http:patch', HTTPPatch)
directives.register_directive('http:options', HTTPOptions)
directives.register_directive('http:head', HTTPHead)
directives.register_directive('http:delete', HTTPDelete)
directives.register_directive('http:copy', HTTPCopy)


class swagger_tag(nodes.Inline, nodes.TextElement):
    pass


class swagger_tag_name(nodes.Inline, nodes.TextElement):
    pass


class swagger_tag_summary(nodes.Inline, nodes.TextElement):
    pass


class SwaggerTag(Directive):

    method = None

    required_arguments = 0
    optional_arguments = 0
    has_content = True
    final_argument_whitespace = True

    option_spec = {
        'synopsis': lambda x: x,
    }

    def run(self):
        node = swagger_tag()
        self.state.nested_parse(self.content, self.content_offset, node)

        # This is the first line of the definition.
        name = node[0].astext()
        node[0].replace_self(swagger_tag_name(name, name))

        # Summary
        summary = self.options.get('synopsis', '')
        node.insert(1, swagger_tag_summary(summary, summary))

        return [node]


directives.register_directive('swagger:tag', SwaggerTag)


class error_writer(object):

    def write(self, line):
        logger.warning(line.strip())


def publish_string(string):
    settings_overrides = {'warning_stream': error_writer()}
    return docutils.core.publish_string(
        string, writer=JSONWriter(),
        settings_overrides=settings_overrides)



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import TestCase

from pecan import set_config
from pecan.testing import load_test_app

__all__ = ['FunctionalTest']


class FunctionalTest(TestCase):
    """Used for functional tests

    Provides a literal application and for testing its integration
    with the framework.
    """
    CONFIG = {
        'server': {
            'port': '8080',
            'host': '0.0.0.0'
        },
        'app': {
            'root': 'fairy_slipper.controllers.root.RootController',
            'modules': ['fairy_slipper'],
            'static_root': '%(confdir)s/public',
            'api_doc': '%(confdir)s/fairy_slipper/tests/api_doc_fixture',
            'template_path': '%(confdir)s/templates',
            'debug': True,
            'errors': {
                '404': '/error/404',
                '__force_dict__': True
            }
        }}

    def __init__(self, *args, **kwargs):
        self.CONFIG = copy.deepcopy(self.CONFIG)
        super(FunctionalTest, self).__init__(*args, **kwargs)

    def setUp(self):
        self.app = load_test_app(copy.deepcopy(self.CONFIG))

    def tearDown(self):
        set_config({}, overwrite=True)



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import os
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from unittest import TestCase
import xml.sax

from fairy_slipper.cmd import docbkx_to_json


class TestChapterParaParser(TestCase):

    def test_para_inline_code(self):
        filename = "test-file.xml"
        test_filename = os.path.dirname(os.path.abspath(__file__))
        test_filename += "/ch_test-v1.xml"

        file_content = """<?xml version="1.0" encoding="UTF-8"?>
        <book xml:id="test-v1" version="1">
        <xi:include href="%s"/>
        </book>
        """ % (test_filename)

        ch = docbkx_to_json.APIRefContentHandler(filename)
        xml.sax.parse(StringIO(file_content), ch)

        self.assertEqual(
            ch.tags,
            [{
                'name': 'test-v1',
                'summary': "Image operations ``show`` all fields.\n"
                "\nCreates, lists, updates, and deletes images.\n"
                "\nCreates, **lists**, ``updates`` and **deletes images**"
                "\n\nCreates, (``x+5``), and deletes **images**."}]
        )

    def test_code_block(self):
        filename = "test-file.xml"
        test_filename = os.path.dirname(os.path.abspath(__file__))
        test_filename += "/ch_test-v3.xml"

        file_content = """<?xml version="1.0" encoding="UTF-8"?>
        <book xml:id="test-v3" version="3">
        <xi:include href="%s"/>
        </book>
        """ % (test_filename)

        ch = docbkx_to_json.APIRefContentHandler(filename)
        xml.sax.parse(StringIO(file_content), ch)

        self.assertEqual(
            ch.tags,
            [{
                'name': 'test-v3',
                'summary': "You can encode sets into a blob. Do something with ``type`` to\n``application/json`` and JSON strings in a ``blob``. Example:\n\n::\n\n   \"blob\": {\n           \"default\": false\n       }\n\nOr:\n\n::\n\n   \"blob\": {\n           \"foobar_user\": [\n               \"role:compute-user\"\n           ]\n       }"  # noqa
            }]
        )

    def test_table_caption(self):
        filename = "test-file.xml"
        test_filename = os.path.dirname(os.path.abspath(__file__))
        test_filename += "/ch_test-table.xml"

        file_content = """<?xml version="1.0" encoding="UTF-8"?>
        <book xml:id="table-v3" version="1">
        <xi:include href="%s"/>
        </book>
        """ % (test_filename)

        ch = docbkx_to_json.APIRefContentHandler(filename)
        xml.sax.parse(StringIO(file_content), ch)

        self.assertEqual(
            ch.tags,
            [{
                'name': 'table-v3',
                'summary': "Creates, lists, updates images.\n"
                "\n**Image status**\n\n++\n"
                "\n**Image with embedded bold status**\n\n++"
            }]
        )

    def test_nested_listitems(self):
        filename = "test-file.xml"
        test_filename = os.path.dirname(os.path.abspath(__file__))
        test_filename += "/ch_test-listitems.xml"

        file_content = """<?xml version="1.0" encoding="UTF-8"?>
        <book xml:id="listitems-v1" version="1">
        <xi:include href="%s"/>
        </book>
        """ % (test_filename)

        ch = docbkx_to_json.APIRefContentHandler(filename)
        xml.sax.parse(StringIO(file_content), ch)

        self.assertEqual(
            ch.tags,
            [{
                'name': 'listitems-v1',
                'summary': "- Para 1, listitem1\n\n  Para 2, listitem1\n\n"
                "  - Embedded item1\n\n  - Embedded item2\n"
                "\n  - Embedded item3\n\n"
                "  Para 3, listitem1\n\n- Para1, listitem2\n"
                "\nsome more para text"
            }]
        )

    def test_link(self):
        filename = "test-file.xml"
        test_filename = os.path.dirname(os.path.abspath(__file__))
        test_filename += "/ch_test-link.xml"

        file_content = """<?xml version="1.0" encoding="UTF-8"?>
        <book xml:id="link-v1" version="1">
        <xi:include href="%s"/>
        </book>
        """ % (test_filename)

        ch = docbkx_to_json.APIRefContentHandler(filename)
        xml.sax.parse(StringIO(file_content), ch)

        self.assertEqual(
            ch.tags,
            [{
                'name': 'link-v1',
                'summary': "To create a keypair, make a "
                "`create keypair"
                "\n<http://developer.openstack.org/#createKeypair>`_"
                " request.\n\n"
                "To test a link that ends the sentence, make a "
                "`create keypair\n"
                "<http://developer.openstack.org/#createKeypair>`_."
            }]
        )



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import unicode_literals

from collections import defaultdict
import json
import logging
from os import path
import re
import six
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

log = logging.getLogger(__name__)

DEFAULT_PORTS = {
    '5000': 'identity',
    '35357': 'identity',
    '8774': 'compute',
    '8776': 'volume',
    '8773': 'compute-ec2',
    '9292': 'image',
    '9696': 'networking',
    '9292': 'image',
    '8082': 'application-catalog',
    '8004': 'orchestration',
    '8080': 'object',
    '8777': 'telemetry',
}

PYTHON_LOG_PREFIX_RE = ("^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{3}) \d+ "
                        "(?P<log_level>[A-Z]+) (?P<logger_name>\S+) "
                        "\[(?P<tags>[^\[\]]+)\] ")

REQUEST_RE = re.compile(PYTHON_LOG_PREFIX_RE + "Request (?P<test>\([^()]+\)):"
                        " (?P<status_code>\d+)"
                        " (?P<method>[A-Z]+) (?P<url>\S+)")

REQUEST1_RE = re.compile(PYTHON_LOG_PREFIX_RE + "Request")

RUBBISH_LINE_RE = re.compile("^    _log_request_full \S+:\d+$")

REQ_RE = re.compile('(req-[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}'
                    '-[a-z0-9]{4}-[a-z0-9]{12})')


class DB(object):

    def __init__(self):
        self.requests = {}
        self.responses = {}

    def _normalize_headers(self, headers):
        return {k.lower(): v for k, v in headers.items()}

    def exists(self, req):
        if req in self.requests:
            return True
        return False

    def create(self, req, request):
        url = urlparse.urlsplit(request['url'])
        port = url.netloc.split(':')[-1]
        service = DEFAULT_PORTS[port]

        self.requests[req] = {
            'service': service,
            'url': urlparse.urlunsplit(('', '') + url[2:]),
            'method': request['method']}
        self.responses[req] = {
            'status_code': request['status_code']}

    def set_request_headers(self, req, headers):
        self.requests[req]['headers'] = self._normalize_headers(headers)

    def set_response_headers(self, req, headers):
        self.responses[req]['headers'] = self._normalize_headers(headers)

    def get_req_or_resp_length(self, req):
        if self.responses[req].get('headers') is not None:
            return int(self.responses[req]['headers'].get('content-length', 0))
        else:
            return int(self.requests[req]['headers'].get('content-length', 0))

    def set_req_or_resp_length(self, req_id, content_length):
        if self.responses[req_id].get('headers'):
            h = self.responses[req_id].get('headers')
            h['content-length'] = content_length
            self.responses[req_id]['headers'] = h
        else:
            if self.requests[req_id].get('headers'):
                h = self.requests[req_id].get('headers')
                h['content-length'] = content_length
                self.requests[req_id]['headers'] = h

    def get_req_or_resp_content_type(self, req):
        if self.responses[req].get('headers') is not None:
            return self.responses[req]['headers'].get('content-type',
                                                      'text/plain')
        else:
            return self.requests[req]['headers'].get('content-type',
                                                     'text/plain')

    def append_req_or_resp_body(self, req_id, string):
        if self.responses[req_id].get('headers') is not None:
            self.responses[req_id]['body'] += string
        else:
            self.requests[req_id]['body'] += string

    def set_req_or_resp_body(self, req_id, string):
        if self.responses[req_id].get('headers') is not None:
            self.responses[req_id]['body'] = string
        else:
            self.requests[req_id]['body'] = string


def parse_logfile(log_file):
    """Yet another shonky stream parser."""
    content_length = 0
    current_req_id = ''
    db = DB()
    for line in log_file:
        if RUBBISH_LINE_RE.match(line):
            continue
        request = REQUEST_RE.match(line)
        if request:
            request_dict = request.groupdict()
            try:
                current_req_id = REQ_RE.match(request_dict['tags']).groups()[0]
            except AttributeError:
                # Swift calls don't have the req tag
                current_req_id = ''
            if current_req_id:
                db.create(current_req_id, request_dict)
        else:
            start_request = REQUEST1_RE.match(line)
            if start_request:
                line = re.sub(PYTHON_LOG_PREFIX_RE, '', line)
                try:
                    current_req_id = REQ_RE.match(
                        start_request.groupdict()['tags']).groups()[0]
                except AttributeError:
                    # Swift calls don't have the req tag
                    current_req_id = ''

                # Skip all boto logs
                if 'boto' == start_request.groupdict()['logger_name']:
                    current_req_id = ''
            try:
                key, value = line.split(':', 1)
            except ValueError:
                # For some wacky reason, when you request JSON,
                # sometimes you get text.  Handle this rad behaviour.
                if not current_req_id:
                    continue

                try:
                    db.append_req_or_resp_body(current_req_id, line)
                except TypeError:
                    log.warning('Failed to find body to add to.')
                continue

            if not current_req_id:
                continue

            if not db.exists(current_req_id):
                log.warning("Can't find request %r" % current_req_id)
                current_req_id = ''
                continue

            key = key.strip()
            value = value.strip()

            if key == 'Request - Headers':
                if current_req_id:
                    db.set_request_headers(current_req_id, eval(value))
            if key == 'Response - Headers':
                if current_req_id:
                    db.set_response_headers(current_req_id, eval(value))
            if key == 'Body':
                if not current_req_id:
                    continue

                content_length = db.get_req_or_resp_length(current_req_id)
                content_type = db.get_req_or_resp_content_type(current_req_id)

                # Trim any messages that are by accident on the end of
                # the line
                if '_log_request_full' in value:
                    value = value.split('_log_request_full')[0]
                if '_check_samples' in value:
                    value = value.split('_check_samples')[0]

                if content_length == 0:
                    body = None
                elif value[:4] == 'None':
                    body = None
                elif 'application/json' in content_type:
                    try:
                        body = json.loads(value)
                    except ValueError:
                        body = value
                        log.warning("Failed to as JSON %r", value)
                        try:
                            body = eval(value)
                            log.warning("Succeed parsing as Python %r", value)
                        except Exception:
                            body = value

                    if not isinstance(body, six.string_types):
                        body = json.dumps(body, indent=2,
                                          sort_keys=True,
                                          separators=(',', ': '))
                else:
                    body = value

                if body is None:
                    content_length = '0'
                else:
                    content_length = str(len(body))
                db.set_req_or_resp_body(current_req_id, body)
                db.set_req_or_resp_length(current_req_id, content_length)

    return db


def main1(log_file, output_dir):
    log.info('Reading %s' % log_file)
    calls = parse_logfile(open(log_file))
    services = defaultdict(list)
    for req in calls.requests:
        call = (calls.requests[req], calls.responses[req])
        services[call[0]['service']].append(call)
    for service, calls in services.items():
        pathname = path.join(output_dir, '%s-examples.json' % (service))
        with open(pathname, 'w') as out_file:
            json.dump(calls, out_file, indent=2)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-v', '--verbose', action='count', default=0,
        help="Increase verbosity (specify multiple times for more)")
    parser.add_argument(
        '-o', '--output-dir', action='store',
        help="The directory to output the JSON files too.")
    parser.add_argument(
        'filename',
        help="File to convert")

    args = parser.parse_args()

    log_level = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG

    logging.basicConfig(
        level=log_level,
        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    filename = path.abspath(args.filename)

    main1(filename, output_dir=args.output_dir)



# Copyright (c) 2015 Russell Sim <russell.sim@gmail.com>
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Server Specific Configurations
server = {
    'port': '8080',
    'host': '0.0.0.0'
}

# Pecan Application Configurations
app = {
    'root': 'fairy_slipper.controllers.root.RootController',
    'modules': ['fairy_slipper'],
    'static_root': '%(confdir)s/public',
    'template_path': '%(confdir)s/fairy_slipper/templates',
    'api_doc': '%(confdir)s/api_doc',
    'debug': True,
    'errors': {
        404: '/error/404',
        '__force_dict__': True
    }
}

logging = {
    'root': {'level': 'INFO', 'handlers': ['console']},
    'loggers': {
        'fairy_slipper': {'level': 'DEBUG', 'handlers': ['console']},
        'pecan.commands.serve': {'level': 'DEBUG', 'handlers': ['console']},
        'py.warnings': {'handlers': ['console']},
        '__force_dict__': True
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'color'
        }
    },
    'formatters': {
        'simple': {
            'format': ('%(asctime)s %(levelname)-5.5s [%(name)s]'
                       '[%(threadName)s] %(message)s')
        },
        'color': {
            '()': 'pecan.log.ColorFormatter',
            'format': ('%(asctime)s [%(padded_color_levelname)s] [%(name)s]'
                       '[%(threadName)s] %(message)s'),
            '__force_dict__': True
        }
    }
}

# Custom Configurations must be in Python dictionary format::
#
# foo = {'bar':'baz'}
#
# All configurations are accessible at::
# pecan.conf
